#| maxdb

A database with fully extensible typesystem.
Data access is based on cells each containing a single value, rather than tables.
Values are identical to references.
Types define the possible values of a cell.
Functions define possible computations: transitions from one value to another
Compound data types, such as structures or relations, may be built using functions.
Initial implementation as simple as possible.
  flat namespace per database.
  very few cell types, mostly just strings
  very few functions, mostly just string concatenation
  flat text file containing all cells only, no transaction log or snapshots
  only one thread may access at a time.
  no background threads: db code loads directly into client application.

Operations here read the entire database linearly, to support
databases that don't fit in RAM.

|#

(in-package :maxdb)

(defclass database ()
  ()
  (:documentation "encapsulates an address space of cells with a common storage and access policy"))

(defclass file-database (database)
  ((filename :initform nil :initarg :filename :accessor db-filename)
   (nextid :initform 0 :initarg :nextid :accessor db-nextid)
   (cache :initform (make-hash-table :test 'equalp) :accessor db-cache)
   (in-transaction :initform nil :accessor in-transaction)
   (xcache :initform (make-hash-table :test 'equalp) :accessor xcache)) ; transaction cache
  (:documentation "a database backed by a file"))

(defmethod initialize-instance :after ((instance file-database) &rest initargs)
  "initialize next cell ID"
  (let ((lastid (lastid instance 'CELL)))
    (setf (db-nextid instance) (+ lastid 1))))

(defun lastid (instance typeid)
  "read through a file database and return the last ID used in a specified type"
  (if (probe-file (db-filename instance))
      (with-open-file (f (db-filename instance) :direction :input)
        (do ((expr (read f nil 'eof) (read f nil 'eof))
             (lastid 0))
            ((eq expr 'eof) lastid)
          (if (equal (symbol-name (car expr)) (symbol-name typeid))
              (setf lastid (cadr expr)))))
    0))

(defmethod assignid ((fdb file-database))
  "assign a new cell ID"
  (let ((id (db-nextid fdb)))
    (incf (db-nextid fdb))
    id))

(defmethod db-filename-temp ((f file-database))
  (filename-temp (db-filename f)))

(defun filename-temp (filename &optional (suffix ".tmp"))
  (concatenate 'string filename suffix))

(defmethod db-filename-transact ((f file-database))
  (filename-temp (db-filename f) ".x"))

(defmethod typename ((ty max:max-type))
  (cond ((max:type-equal ty (max:addr-value max:*string-type*)) 'string)
        ((max:type-equal ty (max:addr-value max:*number-type*)) 'number)))


; do a line-by-line copy of filename to a temporary file
; LINES is a list of strings.
; compares each line to the list of lines specified.
; appends any lines that don't already exist in the file.
; "lines" parameter destructively modified.
; on any error, files are closed and temporary file is left in place for debugging.
; on success, temporary file is moved on top of original file.

; ATOMIC-UPDATE: update the database atomically.
; filename: database filename
; add, update, remove: lists (default empty) of database entries to add, update, or remove.
; entries to add 

(defun write-expr (expr stream)
  (write-char #\( stream)
  (write-string (string-downcase (format nil "~A ~A" (car expr) (cadr expr))) stream)
  (if (null (cddr expr))
      (write-line ")" stream)
    (progn
      ;(format t "expr=~A, (symbolp value)=~A, (stringp value)=~A, (numberp value)=~A~%" expr (symbolp (caddr expr)) (stringp (caddr expr)) (numberp (caddr expr)))
;      (write-string (string-downcase (format nil " ~A)~%" (caddr expr))) stream)
      (write-char #\space stream)
      (write (caddr expr) :case :downcase :stream stream :escape t)
      (write-line ")" stream))))
#|
(defmacro write-expr (expr stream)
  `(progn
     (write ,expr :case :downcase :stream ,stream)
     (terpri)))
|#
(defmacro atomic-debug ()
  `(format t "atomic-debug: expr: ~A, add-values: ~A, update-mappings: ~A, remove-values: ~A, remove-mappings: ~A~%"
           expr
           add-values update-mappings remove-values remove-mappings))
(defun atomic-update (filename &key add-values add-tuples update-mappings remove-values remove-tuples remove-mappings)
;  (declare (optimize (debug 3)))
  (let ((tempfilename (filename-temp filename))
        element)
    (flet ((function-test (x y) 
             ;(format t "function-test: x=~A, y=~A, (equalp (car x)) (car y))=~A, (equalp (cadr x) (cadr y))=~A~%" x y (equalp (car x) (car y)) (equalp (cadr x) (cadr y)))
             (if (and (equalp (symbol-name (car x)) (symbol-name (car y))) (equalp (cadr x) (cadr y)))
                 (progn (setq element y) #| (format t "MATCH~%") |# t)
               nil))
           (tuple-test (m n)
             (and (equalp (symbol-name (car n)) (symbol-name (car m)))
                  (equalp (cdr n) (cdr m)))))
      (with-open-file (fout tempfilename :direction :output :if-exists :error :if-does-not-exist :create)
        (if (probe-file filename) ; ok if no input file
          (with-open-file (fin filename :direction :input)
            (do ((expr (read fin nil 'eof) (read fin nil 'eof)))
                ((eq expr 'eof) "Reached end of file.")
              (if (and (not (member expr remove-values :test #'equalp));if expr is a value in remove-values, skip writing it to file, do not change remove-values
                       (not (member expr remove-mappings :test #'function-test));if expr is a mapping in remove-mappings, skip writing it to file, do not change remove-mappings
                       (not (member expr remove-tuples :test #'tuple-test)))
                  (progn
                    (cond ((member expr add-values :test #'equalp);if expr is a value in add-values, write it to the file and move value to remove-values to clear dupes
                           (setf add-values (delete expr add-values :test #'equalp))
                           (push expr remove-values))
                          ((member expr add-tuples :test #'tuple-test)
                           (setf add-tuples (delete expr add-tuples :test #'tuple-test))
                           (push expr remove-tuples))
                          ((member expr update-mappings :test #'function-test);if expr is a mapping in update-mappings, write updated version to file, move mapping to remove-mappings to clear dupes
                           (setf update-mappings (delete expr update-mappings :test #'function-test))
                           (push expr remove-mappings)
                           (setf (caddr expr) (caddr element))))
                    ;(atomic-debug)
                    (write-expr expr fout))))))
        ;(atomic-debug)
        ; add any remaining types and functions
        (dolist (ty add-values)
          (write-expr ty fout))
        (dolist (tup add-tuples)
          (write-expr tup fout))
        (dolist (func update-mappings) ; add mappings for any function updates we didn't find
          (write-expr func fout))
        (finish-output fout))
      (rename-file tempfilename filename))))

; example: (defparameter globjtype (maxdb:make-type a :id "globj"))
; db: (type globj)

(defclass db-type ()
  ((database :initform nil :initarg :type-database :accessor type-database)
   (id :initform nil :initarg :type-id :accessor type-id)))

;(defun make-type (dbh &optional typeid)
(defmethod make-type (dbh &key id)
   (assert (not (in-transaction dbh)))
  (let ((ty (make-instance 'db-type
                           :type-database dbh
                           :type-id id)))

    (atomic-update (db-filename dbh)
                   :add-values `((type ,id)))

    ty))

;hack
(defparameter *number-type* (make-instance 'db-type :type-id 'number))
(defparameter *string-type* (make-instance 'db-type :type-id 'string))

; example: (defparameter foo (maxdb:type-add-value globjtype :id "foo"))
; db: (globj foo)

(defmethod type-add-value ((ty db-type) &key id)
  (assert (not (in-transaction (type-database ty))))
  (atomic-update (db-filename (type-database ty))
                 :add-values `((,(type-id ty) ,id))))

(defmethod type-remove-value ((ty db-type) &key id)
  (assert (not (in-transaction (type-database ty))))
  (atomic-update (db-filename (type-database ty))
                 :remove-values `((,(type-id ty) ,id))))

; begin transaction- set in-transaction flag
(defmethod begin ((fdb file-database))
  (assert (not (in-transaction fdb)))
;  (if (in-transaction fdb)
;      (rollback fdb))
  (assert (zerop (hash-table-count (xcache fdb))))
  (setf (in-transaction fdb) t))

; rollback- clear xcache, clear in-transaction flag
(defmethod rollback ((fdb file-database))
;  (assert (in-transaction fdb))
  (setf (in-transaction fdb) nil)
  (clrhash (xcache fdb)))

; commit- flush xcache to cache and disk, clear in-transaction flag
(defmethod commit ((fdb file-database))
  (assert (in-transaction fdb))
  (format t "xcache:~%")
  (maphash #'(lambda (key val)
               (format t "~a:~a~%" key val)) (xcache fdb))
  (let ((mappings nil))
    (maphash #'(lambda (key val)
               (remhash key (xcache fdb))
               (setf (gethash key (db-cache fdb)) val)
               (push (list (car key) (cdr key) val) mappings))
           (xcache fdb))
    (atomic-update (db-filename fdb)
                   :update-mappings mappings))

  (setf (in-transaction fdb) nil))

#|
; macro to begin transaction, and rollback in case of any problems
(defmacro transact (dbh &rest args)
  `(block 'maxdb-transaction
     (begin dbh)
     (unwind-protect
         (progn ,@args
                (commit dbh))
       (rollback dbh))))
|#

(defclass db-function ()
  ((database :initform nil :initarg :function-database :accessor function-database)
   (id :initform 0 :initarg :function-id :accessor function-id)))

#| example: (defparameter globj-x (maxdb:make-function a globjtype *number-type* :id "globj-x"))
db: 
(function globj-x)
(domain globj-x globj)
(codomain globj-x number) 
|#
(defmethod make-function ((dbh file-database) (domain db-type) (codomain db-type) &key id)
  (assert (not (in-transaction dbh)))
  (let ((filename (db-filename dbh))
        (func (make-instance 'db-function
                             :function-database dbh
                             :function-id id)))
    (atomic-update filename
                   :add-values `((function ,id))
                   :update-mappings `((domain ,id ,(type-id domain))
                                      (codomain ,id ,(type-id codomain))))
    func))

; example: (maxdb:function-update globj-x foo (max:make-number 0))
; db: (globj-x foo 0)
(defmethod function-update ((func db-function) argument result)
  (let ((db (function-database func))
        (hashkey (cons (function-id func) argument)))
    (if (in-transaction db)
        (setf (gethash hashkey (xcache db)) (max:save-addr result))
      (progn
        (setf (gethash hashkey (db-cache db)) (max:save-addr result))
        (atomic-update (db-filename db)
                     :update-mappings `((,(function-id func) ,argument ,(max:save-addr result))))))))

(defmethod function-remove-mapping ((func db-function) argument)
  (let ((db (function-database func))
        (hashkey (cons (function-id func) argument)))
    (assert (not (in-transaction db)))
    (remhash hashkey (db-cache db))
    (atomic-update (db-filename db)
                   :remove-mappings `((,(function-id func) ,argument)))))

(defmethod check-caches ((dbh file-database) hashkey)
  (if (in-transaction dbh)
      (multiple-value-bind (result foundp)
          (gethash hashkey (xcache dbh))
        (if foundp 
            (values result t)
          (multiple-value-bind (result foundp)
              (gethash hashkey (db-cache dbh))
            (values result foundp))))
    (multiple-value-bind (result foundp)
        (gethash hashkey (db-cache dbh))
      (values result foundp))))

; example: (format t "foo's x coordinate: ~A~%" (maxdb:db-apply a globj-x "foo"))
; output: 0
(defmethod db-apply ((dbh file-database) (func db-function) argument)
  ;(format t "db-apply: func: ~A, arg: ~A~%" (function-id func) argument)
  (let ((hashkey (cons (function-id func) argument)))
    (multiple-value-bind (result foundp)
        (check-caches dbh hashkey)
      (if foundp ; cache hit
          result
        (with-open-file (fin (db-filename dbh))
          (do ((expr (read fin) (read fin nil 'eof)))
              ((eq expr 'eof))
            (if (and (equal (symbol-name (car expr)) 
                            (symbol-name (function-id func)))
                     (equalp (cadr expr) argument))
                (progn
                  (setq result (caddr expr))
                  (setf (gethash hashkey (db-cache dbh)) result)
                  (return-from db-apply result))))
          'UNDEFINED)))))

; db-quotient: 
; return a LISP list of arguments that when applied to the function, produce the given result.
(defmethod quotient ((dbh file-database) (func db-function) result)
  (with-open-file (fin (db-filename dbh))
    (do ((expr (read fin nil 'eof) (read fin nil 'eof))
         (args nil))
        ((eq expr 'eof) args)
      (if (and (equal (symbol-name (car expr))
                      (symbol-name (function-id func)))
               (equalp (caddr expr) result))
          (push (cadr expr) args))))) ; TODO: add mapping to cache?

; return t or nil indicating whether id is a value in type 'ty'.
;(defmethod type-member ((dbh file-database) (ty db-type) id)
