#| relation module for MaxDB

Although functions and relations can be implemented on top of each
other, I'm making both primitive in MaxDB.  While one may be
implemented on top of the other, the API will provide both as
primitive, so there's flexibility to change the implementation later
and the user doesn't have to know the implementation details.

This implementation will provide relations built on top of the
database using similar operations to those used by functions.

|#

#| our example / inspiration for this module is the pipe relation from
the virtual machine monitor.  A pipe is a many to one relation from
sinks to sources, but not every sink has a tuple in this relation, so
the pipe relation isn't a function.  Each sink may have zero or one
source connected at a time.  We need to be able to answer these
questions: 

1. given a sink, what's its currently connected source, or is it
unconnected?

2. given a source, what's the set (or LISP list) of all sinks
depending on it right now?

These are both handled by the selection operator with an ID equality
condition on one of the columns or the other.  We can implement both
by returning a list which would be the empty list, possibly a list of
one source in the case of #1, or possibly a list of many sinks in #2.

Again, the usual strategy is implement it so it works, then speed up
later.  We'll add indices only when they're needed for an application.
They should be far easier to implement once we have the VMM to handle
automatic updating, so we'll try to deal with some slowness until
then.

|#

(in-package :maxdb)

(defclass relation ()
  ((database :initform nil :initarg :relation-database :accessor relation-database)
   (id :initform nil :initarg :relation-id :accessor relation-id)
   (cardinality :initarg :cardinality :accessor relation-cardinality)))

(defmethod make-relation ((dbh file-database) id (ltype db-type) (rtype db-type) &key (cardinality :MANY-TO-MANY))
  (assert (not (in-transaction dbh)))
  (let ((filename (db-filename dbh))
        (rel (make-instance 'relation
                            :relation-database dbh
                            :relation-id id
                            :cardinality cardinality)))
    ; TODO: store types of left and right values.
    ; TODO: Fix this so duplicate IDs aren't added into database.
;    (atomic-update filename
;                   :add-values `((relation ,id)))
    rel))

(defmethod relation-queryl ((rel relation) leftid)
  (let ((filename (db-filename (relation-database rel)))
        result)
    (with-open-file (fin filename)
      (do ((expr (read fin nil 'eof) (read fin nil 'eof)))
          ((eq expr 'eof))
        (if (and (equal (symbol-name (car expr))
                        (symbol-name (relation-id rel)))
                 (equalp (cadr expr) (max:save-addr leftid)))
            (push (caddr expr) result))))
    result))

(defmethod relation-queryr ((rel relation) rightid)
  (let ((filename (db-filename (relation-database rel)))
        result)
    (with-open-file (fin filename)
      (do ((expr (read fin nil 'eof) (read fin nil 'eof)))
          ((eq expr 'eof))
        (if (and (equal (symbol-name (car expr))
                        (symbol-name (relation-id rel)))
                 (equalp (caddr expr) (max:save-addr rightid)))
            (push (cadr expr) result))))
    result))

(defmethod relation-add-tuple ((rel relation) leftid rightid)
  (let ((db (relation-database rel))
        (op (case (relation-cardinality rel)
              (:many-to-many :add-tuples)
              (:many-to-one :update-mappings))))
    (assert (not (in-transaction db)))
    (atomic-update 
     (db-filename db)
     op `((,(relation-id rel) ,(max:save-addr leftid) ,(max:save-addr rightid))))))

(defmethod relation-remove-tuple ((rel relation) leftid rightid)
  (declare (optimize (debug 3)))
  (let ((db (relation-database rel))
        (op (case (relation-cardinality rel) 
              (:many-to-many :remove-tuples) 
              (:many-to-one :remove-mappings))))
    (assert (not (in-transaction db)))
    
    (atomic-update (db-filename db)
                   op `((,(relation-id rel) ,(max:save-addr leftid) ,(max:save-addr rightid))))))
