(in-package max)
(defun make-menu (name items) (cons name items))
(defun menu-name (menu) (car menu))
(defun menu-items (menu) (cdr menu))
(defun make-menu-item (key name help thunk) (list key name help thunk))
(defun menu-item-key (menu-item) (car menu-item))
(defun menu-item-name (menu-item) (cadr menu-item))
(defun menu-item-help (menu-item) (caddr menu-item))
(defun menu-item-thunk (menu-item) (cadddr menu-item))
(defun call-main-menu (main-menu)
 (catch 'end-menu
  (call-menu main-menu :submenup nil)
 )
)
(defun end-menu () (throw 'end-menu t))
(defun previous-menu () (throw 'previous-menu t))
(defun main-menu () (throw 'main-menu t))

(defparameter *quit-item* (make-menu-item
  "Q" "Quit" "Exit from the menu system" #'end-menu)
)
(defparameter *back-item* (make-menu-item
  "B" "Back" "Return to the previous menu" #'previous-menu)
)
(defparameter *mainmenu-item* (make-menu-item
  "M" "Main Menu" "Go to the main menu" #'main-menu)
)
(defun call-menu (menu &key (submenup t) (showheader t) (extratext nil))
  (catch 'previous-menu
    (let* ((extraitems (if submenup 
                              (list nil *back-item* *mainmenu-item* *quit-item*)
                              (list nil *quit-item*))
                     )
         (fullmenu (append menu extraitems))
         (again t))

      (tagbody
        repeat
        (catch (if submenup 'ignore 'main-menu)
          (display-menu fullmenu showheader extratext)
          (display-prompt fullmenu)
          (setq again (execute-choice (menu-items fullmenu) (read-menu-choice fullmenu)))
        )
        (if again (go repeat))
      )
    )
  )
)

(defun display-menu (&optional (menu *main-menu*) (showheader t) (extratext t))
  (if showheader
      (progn
        (terpri)
        (display-menu-header (menu-name menu))
      )
  )
  (if extratext (format t "~%~A" extratext))
  (terpri)
  (display-menu-items (menu-items menu))
  (terpri)
)

(defun display-menu-header (title)
  (let* ((titlerow (format nil "| ~A |" title)) 
         (bar (make-string (length titlerow) :initial-element #\=)))
        (write-line bar)
        (write-line titlerow)
        (write-line bar)
  )
)
(defun display-menu-items (items)
        (dolist (item items)
          (if (not (null item))
              (progn
                (write-string (menu-item-key item))
                (write-string ". ")
                (write-string (menu-item-name item))
                ))
          (terpri)
        )
)

(defun execute-choice (menu-items string)
  (cond ((null menu-items) (write-line (format nil "Unknown option: ~A" string)))
        ((null (car menu-items)) (execute-choice (cdr menu-items) string))
        ((equal (string-downcase string) 
                (string-downcase (menu-item-key (car menu-items))))
         (funcall (menu-item-thunk (car menu-items)))
        )
        (t (execute-choice (cdr menu-items) string))
  )
)
(defun display-prompt (menu) (write-string "--> "))
(defun read-menu-choice (menu) (clear-input) (read-line))

(defun menu-items-from-set (set op)
  (let ((items nil) (n 1))
    (mapset #'(lambda (item)
                 (setq items
                       (cons 
                            (make-menu-item 
                                (format nil "~A" n)
                                (format nil "~A" item)
                                ""   ; No online help
                                #'(lambda () (funcall op item))
                            )
                            items)
                 )
                 (incf n)
              )
            set
    )
    (reverse items)
  )
)

(defun type-menu-item-handler (typeaddr)
  (call-menu (make-menu (format nil "Public interface for ~A" typeaddr) (menu-items-from-set (public-interface typeaddr) #'function-menu-item-handler)))
)

(defparameter *workspace* (get-empty-set *sources*))
(defun source-menu-item-handler (sourceaddr)
    (format t "Performing data request on ~A...~%Result: ~A~%" 
        sourceaddr (source-pull sourceaddr))
    t
)

(defun thunk-to-source (funcaddr)
    (update-workspace (make-function-source funcaddr))
)
(defun update-workspace (sourceaddr)
    (setf *workspace* (set-add *workspace*
        sourceaddr
    ))
    (setf workspace-menu (make-menu "Workspace"
        (menu-items-from-set *workspace* #'source-menu-item-handler)))
)

(defun function-to-source (funcaddr)
 (if (not (catch 'type-mismatch
     (call-menu 
     (make-menu "Select source for argument"
        (menu-items-from-set *workspace* 
           #'(lambda (item) 
               (update-workspace (make-function-source funcaddr item))
               (throw 'previous-menu t)
           )
         )
      )
  )
 ))
    (format t "Type mismatch~%"))
)
(defun function-menu-item-handler (funcaddr)
  (catch 'previous-menu
    (let* ((domain (function-apply *domain* funcaddr))
         (codomain (function-apply *codomain* funcaddr))
         (str (concatenate 'string
              (format nil "Domain          : ~A~%" domain)
              (format nil "Codomain        : ~A~%" codomain)
              (format nil "Implementation:~%~A~%" 
                    (mapping (function-implementation funcaddr)))
              ))
         (menu-items `(,(make-menu-item "1" "Go to domain" "" 
                        (lambda () (type-menu-item-handler domain)))
                       ,(make-menu-item "2" "Go to codomain" "" 
                        (lambda () (type-menu-item-handler codomain)))))
        )
    (call-menu 
        (make-menu (format nil "Function ~A" funcaddr)
                   (append menu-items (list
                       (if (addr-equal domain *null-type*)
                           (make-menu-item "3" "Make source from thunk" ""
                               (lambda () (thunk-to-source funcaddr)))
                           (make-menu-item "3" "Make source from function" "" 
                               (lambda () (function-to-source funcaddr)))
                       )
                   )
        ))
        :extratext str)
    t
    )
  )
)

(defparameter types-menu nil)
(add-init-hook #'(lambda ()
    (setf types-menu 
        (make-menu "Types" 
            (menu-items-from-set 
                (get-public-types) 
                #'type-menu-item-handler)))
))

(defparameter functions-menu nil)
(add-init-hook #'(lambda ()
    (setf functions-menu 
        (make-menu "Functions"
            (menu-items-from-set 
                (get-public-functions) 
                    #'function-menu-item-handler))
    )
))
(defparameter workspace-menu (make-menu "Workspace" NIL))
(defparameter *main-menu* (make-menu
     "Main Menu"
     `(
,(make-menu-item "1" "Types" "Types are abstract collections of values" 
    (lambda () (call-menu types-menu)))
,(make-menu-item "2" "Functions"
  "Pure mathematical functions, with one type for the domain and one for the codomain" 
    (lambda () (call-menu functions-menu)))
,(make-menu-item "3" "Tests" "Run the automated unit tests" (lambda () (test)))
,(make-menu-item "4" "Workspace" "Show the workspace of sources" 
    (lambda () (call-menu workspace-menu)))
     )
))
(defun menu () (call-main-menu *main-menu*))
