(in-package :max)
(defclass max-address ()
  ((addrtype :initarg :addrtype :accessor addr-type)
   (addrvalue :initarg :addrvalue :accessor addr-value)))

(defun make-address-form (typekey addrkey)
        (make-instance 'max-address :addrtype typekey :addrvalue addrkey))
; The first item in the list representing an address is the type.
; Max function ADDR-TYPE
(add-init-hook #'(lambda () 
    (let ((addr-type 
                (make-function 'builtin-function *objects* *types*
                    #'current-type
                )
        ))
        (add-public-function addr-type)
        (add-name addr-type "type")
    )
))
; The second list item in an address is the "value":
; What this actually is depends on the type.
;(defun addr-value (addr)
;        (cadr addr)
;)

(defmethod save-addr ((addr max-address))
        "dispatch on type to save-addr-type"
        (save-addr-type addr (addr-type addr)))
; Two addresses are equal iff:
; their type field is the same symbol and
; their value is equal as defined by the class of the type of the addresses
; Uses addr-value-equal, a generic function which needs to be implemented
; by the class of every type (unless they want addr-equal to fail)
(defun addr-equal (a b)
        (and
                (type-equal (addr-type a) (addr-type b))
                (addr-value-equal 
                        (addr-type a)
                        (addr-value a) 
                        (addr-value b)
                )
        )
)
(add-init-hook #'(lambda ()
    (let ((addr-equal
        (make-function 'builtin-function
            (type-product *objects* *objects*)
            *boolean-type*
            #'(lambda (twoobjs)
                (let ((vals (product-values twoobjs)))
                    (boolean-cond #'(lambda ()
                        (addr-equal (first vals) (second vals))
                    ))
                )
            ))
        ))

        (add-public-function addr-equal)
        (add-name addr-equal "equal")

    )
))

(defclass max-type ()
        nil)
(defmethod factors ((atype max-type))
    (list (make-address-form (addr-value *types*) atype))
)

(defmethod members ((atype max-type))
    (list (make-address-form (addr-value *types*) atype))
)
; Default comparison for two address values is EQ
; since most are now CLOS instances
(defmethod addr-value-equal ((s max-type) a b)
        (eq a b)
)
; the class for a type of types
(defclass types-type (max-type)
        nil
)
; *types-instance* will be the CLOS instance of a type of types.
; There's normally just one such instance.
(defparameter *types-instance* (make-instance 'types-type))

; *types* will be the Max address of the type of types.
(defparameter *types* (make-address-form *types-instance* *types-instance*))
(add-init-hook #'(lambda () (add-name *types* "types")))
; Ignore the key for now, maybe store in a slot for debugging later
(defun make-type (&key (class 'enumerated-type) key)
        (make-address-form (addr-value *types*) (make-instance class))
)
(defun type-equal (a b)
    (or
        (eq a b)
        (and
            (eq (class-of a) (class-of b))
            (type-specific-equal (class-of a) a b)
        )
    )
)
(defun addr-type-equals (type addr)
    (type-equal (addr-value type) (addr-type addr))
)
(defmethod type-specific-equal ((ty max-type) a b)
    (eq a b)
)

(defun typeaddr-instance (typeaddr)
        (addr-value typeaddr)
)
; An enumerated type is created by adding elements one by one,
; explicitly.  It is necessarily finite, and is implemented by a
; hash table: each key corresponds to the address of one item in the 
; type.
; Current examples of enumerated types: *boolean-type*, *types*,
; *functions*
(defclass enumerated-type (max-type)
        ((hash :initform (make-hash-table)
                :accessor hash
                :documentation 
                "Keys are addresses; values are implementation data for each address"
        ))
)
(defun type-hash (typeaddr)
        (hash (typeaddr-instance typeaddr))
)
(defparameter *n_keys* 0)
(defun get-addr-key (sym)
        ; Return key for address using "ADDRnn" if key not specified.
        ; Key names are only for implementation debugging and are not
        ; part of the system design.
        (if (null sym)
                (progn
                        (incf *n_keys*)
                        (nth-value 0
                                (read-from-string
                                        (format nil "~a~a" 'addr *n_keys*)
                                )
                        )
                )
                ; else str argument was provided
                sym
        )
)
#| 
Add a new value to an enumerated type and return the new address.

Required: typeaddr = the max address of the type that the address is
to live in

Optional: key = the symbol to put in the enumerated type's hash table
as this address' key rather than the default ADDRx

Optional: value = value to store in the type's hash table with the
new address as key (Use value only for things that are not Max
addresses and thus can't be returned in functions)
|#
(defun make-address (typeaddr &key key (value nil))
        (let ((type (typeaddr-instance typeaddr))
                (addr-key (get-addr-key key)))

                ; Hash value
                (setf (gethash addr-key (hash type)) value)

                ; Return the new address, including the type it is in
(make-address-form (addr-value typeaddr) addr-key) ) )
; The class of types which are created by taking the sum of other types.
; Note that addresses in sum types are actually in a different type
; depending on which types are in the sum.
; Each sum type has its own instance of class sum-type.
; Two sum types made from the same members will have an identical-looking
;  instance- but won't be the same instance.
; When comparing the addresses of two such sum types, they will be unequal
; in the addr-value-equal sense.  Need a CLASS SUM-TYPE-ADDRESS or something
; if we want to be able to compare them and treated as addr-value-equal.
(defclass sum-type (max-type)
        ((members :reader members :type list
                :initarg :members
                :documentation "The list of types that were summed to make this one"))
)
; Two addresses in the same sum type are equal iff their
; stored addresses are addr-equal.  ADDR-EQUAL compares
; their types and values.  It is assumed that their
; types are one of the member types of this sum type.
(defmethod addr-value-equal ((ss sum-type) a b)
        (addr-equal a b)
)
; Create a type containing the combined
; values from N other types
; (discriminated union?)
; The other types become known as the member types
(defun type-sum (&rest typeaddrs)
        (make-address-form 
                (addr-value *types*) 
                (make-instance 'sum-type :members typeaddrs)
        )
)
; Declare *objects* global that we refer to before it's initialized
(defparameter *objects* nil)
(defun make-sum-constructor (sumtype membertype)
    (if (eq (addr-value sumtype) (addr-value membertype))
        ;then
        (identity-function sumtype)
        ;else
        (make-function 'builtin-function membertype sumtype
                #'(lambda (membervalue)
                        (make-address-form (addr-value membertype) (addr-value membervalue))
                )
        )
    )
)

(add-init-hook #'(lambda ()
  (let ((make-sum-constructor 
            (make-function  'builtin-function 
                            (type-product *types* *types*)
                            *functions*
                #'(lambda (sum-and-member)
                     (apply #'make-sum-constructor (product-values sum-and-member)))
        )))
     (add-public-function make-sum-constructor)
     (add-name make-sum-constructor "constructor")
  )
))

; Given the address of a sum type, return a Max set of
; function addresses, each one of which takes an argument
; in one of the member types and returns an address in the sum type.
(defun get-constructors (sumtype)
        (let    ((funcset (get-empty-set)))

                (map nil #'(lambda (membertype) 
                        (setf funcset (set-add funcset 
                                (make-sum-constructor sumtype membertype)
                        )))
                        (members (addr-value sumtype))
                )

                funcset
        )
)
(add-init-hook #'(lambda ()
    (let ((get-constructors (make-function 'builtin-function *types* *set-type*
            #'get-constructors)))

        (add-public-function get-constructors)
        (add-name get-constructors "constructors")
    )
))
    
; Given an address in a sum type,
; return the address of the type of the current value
; (will be one of the sum type's members at all times)
(defun current-type (sumaddr)
        (make-address-form (addr-value *types*) (addr-type (addr-value sumaddr)))
)

(defun sum-value (sumaddr)
    (addr-value sumaddr)
)



; The class of types created by taking the product of other types.
(defclass product-type (max-type)
        ((factors :reader factors :type list
                :initarg :factors
                :documentation "The list of types that were multiplied to make this one"
        ))
)
; Compare values in the same type product
; Does not check if the factors match the types of the addresses in a or b,
; just if the addresses are equal
;
; Note: This method will only be called if the exact same
; type instance object is used for the type field of the address.
; If you make two set product types from the same factors,
; they will be DIFFERENT TYPES (this will change sometime)
(defmethod addr-value-equal ((ps product-type) a b)
        ; Create list of truth values for comparing each successive address in a and b
        ; then AND them all together (AND returns the list of truth values either way
        (eval (cons 'and (map 'list
                #'(lambda (x y) (addr-equal x y))
                (product-values a) (product-values b))
        ))
)
(defmethod type-specific-equal ((c product-type) a b)
    (eval (cons 'and (map 'list
        #'(lambda (x y) (type-equal x y))
        (factors a) (factors b))
    ))
)

; Create a type containing multiple dimensions:
; one for each of N other types
; Product types
; * Given a list of types, return a new type called a product type.  
;   The original types then become known as the factor types of the product type.
(defun type-product (&rest typeaddrs)
    (if (= 1 (length typeaddrs))
        (car typeaddrs)
        (make-address-form 
                (addr-value *types*) 
                (make-instance 'product-type :factors typeaddrs)
        )
    )
)
; Given the address of a product type, return a list of addresses of its factor types
(defun product-types (prodtype)
        (factors (addr-value prodtype))
)
(defun make-product-addr (prodtype addrlist)
    (if (= 1 (length addrlist))
            (car addrlist)
            (make-address-form (addr-value prodtype) addrlist)
    )
)
(defun product-values (prodaddr)
        (addr-value prodaddr)
)

; Create the Null type, which never has any addresses.
; To be used for the domain of functions with no arguments,
; for example
(defclass null-type (max-type) nil)

(defparameter *null-type* (make-type :class 'null-type))
(add-init-hook #'(lambda () (add-name *null-type* "null")))
