;(declaim (sb-ext:muffle-conditions sb-ext:compiler-note))

(in-package :sync)

; Shortcut to quitting SBCL if you're in this package
#+sbcl (defun quit () (sb-ext:quit))

(defparameter *db* (make-instance 'maxdb:file-database :filename "db.max"))

(defparameter *scene-type* (maxdb:make-type *db* :id 'scene))
(defparameter *scene-name* (maxdb:make-function *db* *scene-type* maxdb:*string-type* :id 'scene-name)) 
(defparameter *gltexttype* (maxdb:make-type *db* :id 'gltext)) 
(defparameter *gltext-string* (maxdb:make-function *db* *gltexttype* maxdb:*string-type* :id 'gltext-string)) 
(defparameter *gltext-scene* (maxdb:make-function *db* *gltexttype* *scene-type* :id 'gltext-scene))
(defparameter *gltext-x* (maxdb:make-function *db* *gltexttype* maxdb:*number-type* :id 'gltext-x)) 
(defparameter *gltext-y* (maxdb:make-function *db* *gltexttype* maxdb:*number-type* :id 'gltext-y)) 
(defparameter *gltext-z* (maxdb:make-function *db* *gltexttype* maxdb:*number-type* :id 'gltext-z)) 

(defparameter *sink-type* (maxdb:make-type *db* :id 'sink))
(defparameter *source-type* (maxdb:make-type *db* :id 'source))
(defparameter *pipe* (maxdb:make-relation *db* 'pipe *sink-type* *source-type* :cardinality :many-to-one))
(defparameter *gltext-sink* (maxdb:make-function *db* *gltexttype* *sink-type* :id 'gltext-sink))
(defparameter *gltext-source* (maxdb:make-function *db* *gltexttype* *source-type* :id 'gltext-source))

;; Frames per second counting

(defstruct fps
  "Frames-per-second counter"
  (frames 0)
  (start-time (get-universal-time)))

(defun incf-fps (fps)
  "Increment FPS counter. When 5s elapsed, print stats and reset counter"
  (let* ((time (get-universal-time))
         (delta (- time (fps-start-time fps))))
    (incf (fps-frames fps))
    (when (>= delta 5)
;      (format t "~&~A frames in ~A seconds = ~A FPS~%"
;              (fps-frames fps) delta (float (/ (fps-frames fps) delta)))
      (setf (fps-frames fps) 0
            (fps-start-time fps) time))
    fps))

;; Fonts

(defparameter +font-filename+ "VeraSe.ttf")
(defparameter +font-ptsize+ 16)
(defparameter +nglyphs+ 128 "Number of glyphs in a font")

(defstruct font
  ; (struct TTF_Font *)  [sic]
  (ttf (ttf:open-font +font-filename+ +font-ptsize+))

  ; array of GL texture names for all glyphs
  (textures (sgum:allocate-foreign-object gl:uint +nglyphs+))

  ; number (not alien) of the display list that draws the first glyph on a quad
  (list-base 0))

(defun power-of-two (input)
  "Round a number up to the nearest power of two"
  (let ((value 1))
    (loop while (< value input)
      do (setq value (* 2 value)))
    value))

;(defparameter *area* (sgum:allocate-foreign-object sdl:rect 1))
; these are always zero (we always blit from the start of the source surface)
;(setf (sdl:rect-x *area*) 0)
;(setf (sdl:rect-y *area*) 0)

(declaim (inline prepare-gl-texture))
(defun prepare-gl-texture (insurf)
  "Create a new surface suitable for use as an OpenGL texture.
   Height and width will be powers of two, min size 64x64"
  (let* ((w (max 64 (power-of-two (sdl:surface-w insurf))))
         (h (max 64 (power-of-two (sdl:surface-h insurf))))
         ; assumes little endian
         (outsurf (sdl:create-rgb-surface sdl:+swsurface+ w h 32 #x000000FF #x0000FF00 #x00FF0000 #xFF000000))
         (pixel-format (sdl:surface-format outsurf)))

    (sgum:with-foreign-objects ((area sdl:rect))

      ; define area to blit from
      (setf (sdl:rect-x area) 0)
      (setf (sdl:rect-y area) 0)
      (setf (sdl:rect-w area) (sdl:surface-w insurf))
      (setf (sdl:rect-h area) (sdl:surface-h insurf))

      ;(sdl:fill-rect outsurf *area* (sdl:map-rgba pixel-format 0 0 0 0))

      (sdl:set-alpha insurf 0 255)

      (sdl:blit-surface insurf area outsurf area)

      outsurf)))

(defparameter *white* (sgum:allocate-foreign-object sdl:color 1))
(setf (sdl:color-r *white*) 255)
(setf (sdl:color-g *white*) 255)
(setf (sdl:color-b *white*) 255)
(defparameter *black* (sgum:allocate-foreign-object sdl:color 1))
(setf (sdl:color-r *black*) 0)
(setf (sdl:color-g *black*) 0)
(setf (sdl:color-b *black*) 0)
(defparameter *red* (sgum:allocate-foreign-object sdl:color 1))
(setf (sdl:color-r *red*) 255)
(setf (sdl:color-g *red*) 0)
(setf (sdl:color-b *red*) 0)
(defparameter *gray* (sgum:allocate-foreign-object sdl:color 1))
(setf (sdl:color-r *gray*) 128)
(setf (sdl:color-g *gray*) 128)
(setf (sdl:color-b *gray*) 128)
(defparameter *green* (sgum:allocate-foreign-object sdl:color 1))
(setf (sdl:color-r *green*) 0)
(setf (sdl:color-g *green*) 255)
(setf (sdl:color-b *green*) 0)


(defparameter *char-advance* (sgum:allocate-foreign-object gl:int 1))
(defparameter *char-minx* (sgum:allocate-foreign-object gl:int 1))
(defparameter *char-maxx* (sgum:allocate-foreign-object gl:int 1))
(defparameter *char-miny* (sgum:allocate-foreign-object gl:int 1))
(defparameter *char-maxy* (sgum:allocate-foreign-object gl:int 1))

(defmacro printsurf (a)
    `(format t "~A~%" (loop for row from 0 to (- (sdl:surface-h ,a) 1) 
      collect (loop for col from 0 to (- (sdl:surface-w ,a) 1) 
                collect (format nil "~X" (sdl:get-pixel ,a col row))))))

(defun make-dlist (font ch)
  "Create a display list of a quad, texture-mapped with a single character"
  (let* ((texture-number (sgum:deref-array (font-textures font) gl:uint ch))
         ;(glyphsurf (ttf:render-glyph-blended (font-ttf font) ch (sgum:deref-pointer *white* sdl:color)))
         ; WORKS with my patch converting SDL_Color to a uint32:
         (glyphsurf (ttf:render-glyph-blended (font-ttf font) ch #x00FFFFFF)) ; #x00BBRRGG
         ; TRY NEW Method with individual a,b,g,r: fails.
         ;(glyphsurf (ttf:render-glyph-blended (font-ttf font) ch #xFF #xFF #xFF #xFF)) ; #?? #nothing #?? #??
         ;(glyphsurf (ttf:render-glyph-shaded (font-ttf font) ch (sgum:deref-pointer *white* sdl:color) (sgum:deref-pointer *black* sdl:color)))
         ;(glyphsurf (ttf:render-glyph-solid (font-ttf font) ch (sgum:deref-pointer *white* sdl:color)))
         (char-width (sdl:surface-w glyphsurf))
         (char-height (sdl:surface-h glyphsurf))
         (surf (prepare-gl-texture glyphsurf))
         (tex-width (sdl:surface-w surf))
         (tex-height (sdl:surface-h surf))
         ; x and y are texture coordinates (0 .. 1)
         ; for finding input surface within the new output surface.
         ; Use these in your call to gl:tex-coord-2f (the other corner is 0,0)
         (texmaxx (coerce (/ char-width tex-width) 'float))
         (texmaxy (coerce (/ char-height tex-height) 'float))
         )
    (format t "dlist #~A: tex#~A char-width: ~A char-height: ~A tex-width: ~A tex-height: ~A texmaxx: ~A texmaxy: ~A~%"
            ch texture-number char-width char-height tex-width tex-height texmaxx texmaxy)

    ;(sdl:save-bmp glyphsurf (format nil "glyph-~A.bmp" ch))
    ; Create the texture from the character image
    (gl:bind-texture gl:+texture-2d+ texture-number)
    (gl:tex-parameter-i gl:+texture-2d+ gl:+texture-min-filter+ gl:+linear+)
    (gl:tex-parameter-i gl:+texture-2d+ gl:+texture-mag-filter+ gl:+linear+)
    ;(gl:tex-image-2d gl:+texture-2d+ 0 gl:+rgba+ tex-width tex-height 0 gl:+luminance-alpha+ gl:+unsigned-byte+ (sdl:surface-pixels surf))
    ;(printsurf surf)
    ;(sdl:save-bmp surf (format nil "prep-~A.bmp" ch))
    (gl:tex-image-2d gl:+texture-2d+ 0 4 tex-width tex-height 0 gl:+rgba+ gl:+unsigned-byte+ (sdl:surface-pixels surf))
    ;(gl:tex-image-2d gl:+texture-2d+ 0 3 tex-width tex-height 0 gl:+rgba+ gl:+unsigned-byte+ (sdl:surface-pixels surf))
    ;(gl:tex-image-2d gl:+texture-2d+ 0 gl:+rgba+ tex-width tex-height 0 gl:+rgba+ gl:+unsigned-byte+ (sdl:surface-pixels surf))

    ; Create a display list that draws a quad texture-mapped with that character
    (gl:new-list (+ ch (font-list-base font)) gl:+compile+)
    (gl:bind-texture gl:+texture-2d+ texture-number)
    (gl:push-matrix)
    (ttf:glyph-metrics (font-ttf font) ch *char-minx* *char-maxx* *char-miny* *char-maxy* *char-advance*)
    ;(gl:translate-f 0f0 (- (coerce char-height 'single-float)) 0f0)
    ;(gl:translate-f (coerce (sgum:deref-pointer *char-minx* gl:int) 'single-float)
    ;                (coerce (- (ttf:font-ascent (font-ttf font)) (sgum:deref-pointer *char-miny* gl:int)) 'single-float)
    ;                0f0)
    (gl:translate-f (coerce (sgum:deref-pointer *char-minx* gl:int) 'single-float)
                    (coerce (- (sgum:deref-pointer *char-maxy* gl:int)) 'single-float)
                    0f0)
    (gl:color-4f 1f0 1f0 1f0 1f0)
    (gl:begin gl:+quads+)
      (gl:tex-coord-2f 0f0 0f0)
      (gl:vertex-2i 0 0)
      (gl:tex-coord-2f 0f0 texmaxy)
      (gl:vertex-2i 0 char-height)
      (gl:tex-coord-2f texmaxx texmaxy)
      (gl:vertex-2i char-width char-height)
      (gl:tex-coord-2f texmaxx 0f0)
      (gl:vertex-2i char-width 0)
    (gl:end)
    (gl:pop-matrix)
    (gl:translate-f (coerce (sgum:deref-pointer *char-advance* gl:int) 'single-float) 0f0 0f0)
    (gl:end-list)
  )
)

(defun init-font ()
  (let (font)
    (ttf:init)
    (setf font (make-font))
    (gl:gen-textures +nglyphs+ (font-textures font))
    (setf (font-list-base font) (gl:gen-lists +nglyphs+))
    (loop for i from 0 to (- +nglyphs+ 1) do (make-dlist font i))
    font
  )
)

(defparameter *x* 0.0f0)
(defparameter *y* 0.0f0)
(defparameter *z* -6.0f0)

(defparameter *rtri* 0f0)
(defparameter *drtri* 0.2f0)
(defparameter *rquad* 0f0)
(defparameter *drquad* 0.15f0)

; Delta for keyboard scroll
(defparameter *dx* 0.5f0)
(defparameter *dy* 0.5f0)
(defparameter *dz* 0.5f0)
; Delta for mouse scroll
(defparameter *dxm* 0.04f0)
(defparameter *dym* -0.04f0) ; negative for invert mouse
;Rotational position around the x and y axes
(defparameter *xr* 0f0)
(defparameter *yr* 0f0)
; Delta for mouse rotate
(defparameter *dxr* 0.5f0)
(defparameter *dyr* 0.5f0)

; Current mouse position
(defparameter *mx* 0)
(defparameter *my* 0)

; Triangle color
(defparameter *trir* 1f0)
(defparameter *trib* 0f0)
(defparameter *trig* 0f0)
; Square color
(defparameter *quadr* 0f0)
(defparameter *quadb* 1f0)
(defparameter *quadg* 0f0)

(defparameter *font-r* 1f0)
(defparameter *font-g* 1f0)
(defparameter *font-b* 1f0)



;(defparameter *max-key* 0)
; 319 is the highest key code returned by SDL based on experimentation
(defparameter *keys* (make-array sdl:+k-last+ :initial-element nil))
; Mouse buttons
(defparameter *mb* (make-array 7 :initial-element nil))

(defparameter *rp* (sgum:allocate-foreign-object gl:float 4)) ; raster position
(defparameter *rpv* (sgum:allocate-foreign-object gl:boolean)) ; raster position validity


(defparameter *select* nil)
(defparameter *projection-base* (sgum:allocate-foreign-object gl:float 16))
(defparameter *proj-presel* (sgum:allocate-foreign-object gl:double 16))

(defparameter +selection-buffer-len+ 512)

(defun select (surface width height font)
  ; ARGH! with-foreign-object fails, I give up, use hard to read method that works:
  (let ((buffer (sgum:allocate-foreign-object gl:uint +selection-buffer-len+)))
    (unwind-protect
        (let ((viewport (sgum:allocate-foreign-object gl:int 4)))
          (unwind-protect
              (let ((hits 0))
                (gl:get-integerv gl:+viewport+ viewport)
                ;(format t "Viewport: ~A~%" (loop for n from 0 to 3 collect (sgum:deref-array viewport gl:int n)))
                (gl:select-buffer +selection-buffer-len+ buffer)
                (gl:render-mode gl:+select+)
                (gl:init-names)
                (gl:push-name 0)
                (gl:matrix-mode gl:+projection+)
                (gl:push-matrix)
                (gl:get-doublev gl:+projection-matrix+ *proj-presel*)
                (gl:load-identity)
                (glu:pick-matrix (coerce *mx* 'double-float)
                                 (coerce (- (sgum:deref-array viewport gl:int 3) *my*) 'double-float)
                                 1d0 1d0 viewport)
                (gl:get-floatv gl:+projection-matrix+ *projection-base*)
                (format t "Pick matrix:~%")
                (print-4x4-matrix *projection-base*)
                (glu:perspective 45d0
                                 (coerce
                                  (/ (- (sgum:deref-array viewport gl:int 2)
                                        (sgum:deref-array viewport gl:int 0))
                                     (- (sgum:deref-array viewport gl:int 3)
                                        (sgum:deref-array viewport gl:int 1)))
                                  'double-float)
                                 0.1d0
                                 100.0d0)
                (gl:matrix-mode gl:+modelview+)
                (draw-scene *scene* surface width height font)
                (setq nhits (gl:render-mode gl:+render+))
                ;(format t "Hits: ~A~%" hits)
                (setf *select*
                      (if (> nhits 0)
                          ; get all the names for just the first hit
                          (loop for i from 0 to (- (sgum:deref-array buffer gl:uint 0) 1)
                            collect (sgum:deref-array buffer gl:uint (+ 3 i)))

                        ; clicked on empty space, clear selection
                        nil))

                (format t "Select: ~A~%" *select*)

                (gl:matrix-mode gl:+projection+)
                (gl:load-identity)
                (gl:get-floatv gl:+projection-matrix+ *projection-base*)
                (gl:pop-matrix)
                (gl:matrix-mode gl:+modelview+))

            (sgum:free-foreign-object viewport)))
        (sgum:free-foreign-object buffer))))

(defun move ()
;  (if (= key (char-code #\a))
;      (setf *x* (- *x* *dx*)))
;  (if (= key (char-code #\d))
;      (setf *x* (+ *x* *dx*)))
;  (if (= key (char-code #\w))
;      (setf *y* (+ *y* *dy*)))
;  (if (= key (char-code #\s))
;      (setf *y* (- *y* *dy*)))
  (if (aref *keys* (char-code #\a))
      (decf *x* *dx*))
  (if (aref *keys* (char-code #\d))
      (incf *x* *dx*))
  (if (aref *keys* (char-code #\w))
      (incf *y* *dy*))
  (if (aref *keys* (char-code #\s))
      (decf *y* *dy*))
)

(defun resize (surface width height)
;  (format t "resize ~A ~A~%" width height)
  (when (zerop height)
    (setf height 1))
  (let ((ratio (coerce (/ width height) 'double-float)))
    (gl:viewport 0 0 width height)
    (gl:matrix-mode gl:+projection+)
    (gl:load-identity)
    (gl:get-floatv gl:+projection-matrix+ *projection-base*)
    (glu:perspective 45.0d0 ratio 0.1d0 100.0d0)
    (gl:matrix-mode gl:+modelview+)
    (gl:load-identity)
    t))

(defun enter-2d-mode (width height)
  (gl:push-attrib (logior gl:+enable-bit+ gl:+transform-bit+))
;  (gl:push-attrib gl:+enable-bit+)
  ;(gl:disable gl:+depth-test+)
  (gl:disable gl:+cull-face+)
  (gl:enable gl:+texture-2d+)
    
  ; This allows alpha blending of 2D textures with the scene
  (gl:enable gl:+blend+)
  (gl:blend-func gl:+src-alpha+ gl:+one-minus-src-alpha+)

  ; this is called automatically on resize event
  ;(gl:viewport 0 0 width height)

  (gl:matrix-mode gl:+projection+)
  (gl:push-matrix)
  (gl:load-identity)
  (gl:mult-matrix-f *projection-base*)
  ;(format t "Entering 2D Mode.  Pick Matrix is:~%")
  ;(print-4x4-matrix *projection-base*)

  (gl:ortho 0d0 (coerce width 'double-float)
            (coerce height 'double-float)
            0d0 0d0 1d0)

  (gl:matrix-mode gl:+modelview+)
  (gl:push-matrix)
  (gl:load-identity)

  (gl:tex-env-i gl:+texture-env+ gl:+texture-env-mode+ gl:+modulate+)
)

(defun leave-2d-mode ()
  (gl:matrix-mode gl:+projection+)
  (gl:pop-matrix)

  (gl:matrix-mode gl:+modelview+)
  (gl:pop-matrix)

  (gl:pop-attrib)
)

(defun gl-print (text font)
  (let ((arr (sgum:allocate-foreign-object gl:ubyte (+ (length text) 1))))
    (unwind-protect
        (progn
          (loop for i from 0 to (- (length text) 1) do (setf (sgum:deref-array arr gl:ubyte i) (char-code (aref text i))))
          (setf (sgum:deref-array arr gl:ubyte (length text)) 0) ;Null terminated
          (gl:push-attrib (logior gl:+list-bit+ gl:+current-bit+ gl:+enable-bit+ gl:+transform-bit+))
          (gl:matrix-mode gl:+modelview+)
          (gl:disable gl:+lighting+)
          (gl:enable gl:+texture-2d+)
          ;(gl:disable gl:+depth-test+)
          (gl:enable gl:+blend+)
          ;(gl:disable gl:+blend+)
          ;(gl:blend-func gl:+src-alpha+ gl:+one-minus-src-alpha+)
          ;(gl:blend-func gl:+one+ gl:+zero+)
          (gl:blend-func gl:+src-alpha+ gl:+one+)
          ;(gl:blend-func gl:+one-minus-src-alpha+ gl:+src-alpha+)
          (gl:color-4f 1f0 1f0 1f0 1f0)
          ;(gl:translate-f 0f0 0f0 -0.11f0)
          (gl:list-base (font-list-base font))
          (gl:call-lists (length text) gl:+unsigned-byte+ arr)
          (gl:pop-attrib))
      (sgum:free-foreign-object arr))))

(defstruct color (r 0f0) (g 0f0) (b 0f0) (a 1f0))

;non-draggable widget for arrows
(defclass glwidget ()
  ((color :initarg :color :accessor color)
   (primitive :initarg :primitive :accessor primitive)
   (selected :initform nil :accessor selected)))

(defclass globj (glwidget)
  ((dragx :initform nil :accessor dragx :documentation "x distance from object to dragging cursor - world coordinates")
   (dragy :initform nil :accessor dragy :documentation "y distance from object to dragging cursor - world coordinates")
   ;dragz not stored, since we don't move objects within z dimension yet
   ))

(defmethod draggable ((o globj)) t)

; coordinates of cursor at selection time - screen coordinates - use for drag threshold
(defparameter *clickx* nil)
(defparameter *clicky* nil)

(defclass glpointable ()
  ((arrows :initform nil :accessor obj-arrows)))

(defclass glarrow (glwidget)
  ;;head, tail, coordinates determined dynamically from attachment points.
  ((headobj :initarg :headobj :initform nil :accessor headobj)
   (tailobj :initarg :tailobj :initform nil :accessor tailobj)
   (head-source :initarg :head-source :initform nil :accessor head-source)
   (tail-sink :initarg :tail-sink :initform nil :accessor tail-sink)))
   ;(head-source-id :initarg :head-source-id :initform nil :accessor head-source-id)
   ;(tail-sink-id :initarg :tail-sink-id :initform nil :accessor tail-sink-id)))

(defmethod draggable ((o glarrow)) nil)

(defmethod obj-delete-arrow ((p glpointable) (a glarrow))
  (setf (obj-arrows p) (delete a (obj-arrows p))))

(defclass glsource (maxvmm:source)
  ((id :initarg :id :initform nil :accessor source-id)))

(defclass glsink (maxvmm:sink)
  ((id :initarg :id :initform nil :accessor sink-id)
   (parent :initarg :parent :initform nil :accessor sink-parent)))

(defclass funcom-sink (glsink)
  ((value :initarg :value :initform nil :accessor sink-value))
  (:documentation "sink in a functional component"))

#|(defsetf 
(defmethod sink-value ((fcs funcom-sink))
  "fake reader method - call source on nil slot"
  (if (null (slot-value fcs 'value))
      (let ((mysource (sink-source fcs)))
        (if (not (null mysource))
            (setf (sink-value fcs) (source-value mysource)))))
  (slot-value fcs 'value))|#

(defmethod obj-delete :before ((p glpointable))
    (scene-delete-arrows *scene* (obj-arrows p)))

(defun make-arrow (tail head)
  (let ((sink (obj-sink tail))
        (source (obj-source head)))
    (if (not (and sink source))
        (return-from make-arrow)

      (let* ((blue (make-color :b 1f0))
             (a (make-instance 'glarrow :color blue 
                               :tailobj tail 
                               :headobj head
                               :tail-sink sink
                               :head-source source)))
        (format t "make-arrow adding arrow~%")
        (maxvmm:make-pipe sink source)
        (maxdb:relation-add-tuple *pipe* (max:make-number (sink-id sink)) (max:make-number (source-id source)))
        (push a (obj-arrows tail))
        (push a (obj-arrows head))
        (scene-add-arrow *scene* a)))))

;(defmethod arrow-head-source ((a glarrow))

(defun focussed (obj)
  (eq *keyfocus* obj))

(defclass scene ()
  ((id :initarg :id :initform nil :accessor scene-id)
   (shapes :initarg :shapes :initform nil :accessor shapes)
   (texts :initarg :texts :initform nil :accessor texts)
   (arrows :initarg :arrows :initform nil :accessor scene-arrows)))

(defmethod drawlist ((s scene))
  (append (shapes s) (texts s) (scene-arrows s)))

(defmethod scene-add-shape ((s scene) (obj globj))
  (push obj (shapes s)))

(defmethod scene-add-arrow ((s scene) (a glarrow))
  (push a (scene-arrows s)))

(defgeneric draw (obj width height))

(defclass glshape (globj)
  ((vertices :initarg :vertices :accessor vertices)))

(defmethod draw ((obj glwidget) width height)
  (gl:color-4f (color-r (color obj)) (color-g (color obj)) (color-b (color obj)) (color-a (color obj))))

(defmethod draw ((obj glshape) width height)
  (call-next-method)
  (gl:translate-f (obj-x obj) (obj-y obj) (obj-z obj))
  (unless (null (vertices obj))
    (progn
      (gl:begin (primitive obj))
      (mapc #'(lambda (coords) (apply #'gl:vertex-3f coords)) (vertices obj))
      (gl:end))))

(defmacro radians2degrees (rad) `(* 180 (/ ,rad PI)))

(defmethod draw ((a glarrow) width height)
  (call-next-method)
  (if (selected a)
      (gl:color-4f 1f0 1f0 1f0 1f0))
  (let* ((headx (obj-x (headobj a)))
         (heady (obj-y (headobj a)))
         (headz (obj-z (headobj a)))
         (tailx (obj-x (tailobj a)))
         (taily (obj-y (tailobj a)))
         (tailz (obj-z (tailobj a)))
         (rotx (- headx tailx))
         (roty (- heady taily))
         (len (sqrt (+ (expt rotx 2) (expt roty 2))))
         (angle (radians2degrees (asin (/ rotx len)))))
    (gl:translate-f 0f0 0f0 0f0)
    (gl:begin gl:+lines+)
      (gl:vertex-3f headx heady headz)
      (gl:vertex-3f tailx taily tailz)
    (gl:end)
    (gl:push-matrix)
    (gl:translate-f headx heady headz)
    (gl:rotate-d angle 0d0 0d0 -1d0)
    (gl:begin gl:+triangles+)
      (gl:vertex-3f 0f0 0f0 0f0)
      (gl:vertex-3f -0.05f0 -0.1f0 0f0)
      (gl:vertex-3f 0.05f0 -0.1f0 0f0)
    (gl:end)
    (gl:pop-matrix)

))

(defun invert-colors (color)
  (setf (color-r color) (- 1f0 (color-r color)))
  (setf (color-g color) (- 1f0 (color-g color)))
  (setf (color-b color) (- 1f0 (color-b color))))

(defparameter *drag-threshold* 3) ; minimum pixel distance to begin object drag
; 150ms default double-click sensitivity (3/20 second)
;(defparameter *double-click-sensitivity* (/ (* 3 internal-time-units-per-second) 20))
(defparameter *double-click-sensitivity* 350)
; set last click back in time far enough to make it impossible to trigger double-click right away.
(defparameter *last-click-time* (- (get-internal-real-time) *double-click-sensitivity*))

(defmethod handle-select ((a glarrow) subnames x y)
  (setf (selected a) t))

(defmethod handle-select ((obj globj) subnames x y)
  (let* ((curtime (get-internal-real-time))
         (delta (- curtime *last-click-time*)))
    (if (< delta *double-click-sensitivity*)
        (start-keyfocus obj))
        ;(format t "Double-click received!  Delta ~A~%" delta))
    (setf *last-click-time* curtime)
    (setf (selected obj) t)
    (if (draggable obj)
        (multiple-value-bind (tclickx tclicky tclickz) ;translated click coordinates
            (position-object x y)
          (setf (dragx obj) (- tclickx (obj-x obj))) ; x position of pointer on object when drag starts, relative to objx
          (setf (dragy obj) (- tclicky (obj-y obj)))
          (setf *clickx* x)
          (setf *clicky* y)))))

(defmethod handle-select ((obj glshape) subnames x y)
  (invert-colors (color obj))
  (call-next-method))

(defclass gltext (globj glpointable)
  ((id :initform nil :initarg :id :accessor gltext-id)
   (height :initform nil :accessor height)
   (width :initform nil :accessor width)
   (point :initform 0 :accessor point :documentation "cursor pos is left of the character at index point")
   (vmsource :initarg :source :accessor vmsource)
   (vmsink :initarg :sink :accessor vmsink)))

(defmethod text ((tobj gltext))
  (maxdb:db-apply *db* *gltext-string* (gltext-id tobj)))

(defun text-update (tobj value)
  (maxdb:function-update *gltext-string* (gltext-id tobj) (max:make-string-address value))
  (unless (focussed tobj) (maxvmm:update (vmsource tobj) value)))

(defsetf text text-update)

(defmethod obj-x ((tobj gltext))
  (maxdb:db-apply *db* *gltext-x* (gltext-id tobj)))
(defmethod obj-x-update ((tobj gltext) x)
  (maxdb:function-update *gltext-x* (gltext-id tobj) (max:make-number x)))
(defsetf obj-x obj-x-update) ; also used by concat.lisp

(defmethod obj-y ((tobj gltext))
  (maxdb:db-apply *db* *gltext-y* (gltext-id tobj)))
(defmethod obj-y-update ((tobj gltext) y)
  (maxdb:function-update *gltext-y* (gltext-id tobj) (max:make-number y)))
(defsetf obj-y obj-y-update)

(defmethod obj-z ((tobj gltext))
  (maxdb:db-apply *db* *gltext-z* (gltext-id tobj)))
(defmethod obj-z-update ((tobj gltext) z)
  (maxdb:function-update *gltext-z* (gltext-id tobj) (max:make-number z)))
(defsetf obj-z obj-z-update)

(defmethod scene-add-text ((s scene) (text gltext))
  (push text (texts s)))

(defmethod set-text ((obj gltext) text)
        (setf (text obj) text))

(defmethod handle-select ((obj gltext) subnames x y)
  (setf (point obj)
        (if (null subnames)
            ;clicked near the text, but not on any character
            (length (text obj))
          (car subnames)))
  (format t "Set point to ~A~%" (point obj))
  (call-next-method))

(defmethod draw ((obj gltext) width height)
  (let ((projmatrix *proj-presel*)
        (modelmatrix *mvm*)
        (viewport (sgum:allocate-foreign-object gl:int 4))
        (winx (sgum:allocate-foreign-object gl:double 1))
        (winy (sgum:allocate-foreign-object gl:double 1))
        (winz (sgum:allocate-foreign-object gl:double 1))
        x y z)

    (unwind-protect
        (progn
          ;(gl:get-doublev gl:+projection-matrix+ projmatrix)
          ;(gl:get-doublev gl:+modelview-matrix+ modelmatrix)
          (gl:get-integerv gl:+viewport+ viewport)

          (glu:project (coerce (obj-x obj) 'double-float)
                       (coerce (obj-y obj) 'double-float)
                       (coerce (obj-z obj) 'double-float)
                       modelmatrix
                       projmatrix
                       viewport
                       winx winy winz)

          (setf x (sgum:deref-pointer winx gl:double))
          (setf y (sgum:deref-pointer winy gl:double))
          (setf z (sgum:deref-pointer winz gl:double))

          ;(format t "projected text to: <~A,~A,~A>~%" x y z)

          ; round the X,Y coordinates to nearest pixel, for crisp, bright text
          ; need to still be a floating point though
          (setf x (coerce (round x) 'double-float))
          (setf y (coerce (round y) 'double-float))

          ;(gl:translate-f (pos-x (pos obj)) (pos-y (pos obj)) (pos-z (pos obj)))
          ;(unless (= 0f0 *z*)
          ;  (gl:raster-pos-3f 0f0 0f0 0f0))

          ;(gl:get-floatv gl:+current-raster-position+ *rp*)

          (enter-2d-mode width height)
          ;(gl:color-4f (color-r (color obj)) (color-g (color obj)) (color-b (color obj)) (color-a (color obj)))
          (gl:enable gl:+texture-2d+)
          ;(gl:translate-f (sgum:deref-array *rp* gl:float 0)
          ;                (- height (raster-y))
          ;                0f0)
          (gl:translate-d x (- height y) 0d0)
          (gl:push-matrix)
          ;(gl-print (text obj) (font obj))
          (gl:push-attrib (logior gl:+list-bit+ gl:+current-bit+ gl:+enable-bit+ gl:+transform-bit+))
          ;(gl:matrix-mode gl:+modelview+)
          (gl:disable gl:+lighting+)
          (gl:enable gl:+blend+)
          (gl:blend-func gl:+src-alpha+ gl:+one+)
          (gl:color-4f 1f0 1f0 1f0 1f0)
          ;(gl:call-lists (length text) gl:+unsigned-byte+ arr)
          (let* ((text (text obj))
                 ;(font (font obj))
                 (font *font*)
                 (width 0)
                 (height (ttf:font-height (font-ttf font)))
                 (char-advance (sgum:allocate-foreign-object gl:int 1))
                 (ascent (ttf:font-ascent (font-ttf font)))
                 (top (- ascent))
                 (bottom (+ top height)))
            (unwind-protect
                (progn
            ;(setf (sgum:deref-pointer char-advance gl:int) 0)
            (loop for i from 0 to (length text) do
              (gl:disable gl:+texture-2d+)
              (if (and (focussed obj) (= i (point obj)) (cursor-visible))
                  ; draw cursor here
                  (progn
                    (gl:color-4f 1f0 1f0 1f0 1f0)
                    (gl:begin gl:+lines+)
                      (gl:vertex-2i 0 bottom)
                      (gl:vertex-2i 0 top)
                    (gl:end)))
              when (<= i (- (length text) 1)) do
              (let ((ch (char-code (aref text i))))
                (ttf:glyph-metrics (font-ttf font) ch nil nil nil nil char-advance)
                (incf width (sgum:deref-pointer char-advance gl:int))
                ; draw invisible box around each half of the letter so clicking that half moves point
                (gl:color-4f 0f0 0f0 0f0 0f0)
                (let* ((full (sgum:deref-pointer char-advance gl:int))
                       (half (round (/ full 2))))
                  ;(format t "top: ~A bottom: ~A~%" top bottom)
                  (gl:push-name i)
                  (gl:begin gl:+quads+)
                    (gl:vertex-2i 0 bottom)
                    (gl:vertex-2i 0 top)
                    (gl:vertex-2i half top)
                    (gl:vertex-2i half bottom)
                  (gl:end)
                  (gl:load-name (+ i 1))
                  (gl:begin gl:+quads+)
                    (gl:vertex-2i half bottom)
                    (gl:vertex-2i half top)
                    (gl:vertex-2i full top)
                    (gl:vertex-2i full bottom)
                  (gl:end))
                (gl:enable gl:+texture-2d+)
                (gl:call-list (+ ch (font-list-base font)))
                (gl:pop-name)))
            ;    (format t "Text width: ~A, Height: ~A~%" width height)
            (setf (width obj) width)
            (setf (height obj) height)

            (gl:pop-attrib)
            (gl:pop-matrix)
            (gl:disable gl:+texture-2d+)
            (if (selected obj)
                (progn
                  (gl:color-4f 1f0 1f0 1f0 1f0)
                  (gl:begin gl:+line-loop+)
                    (gl:vertex-2i 0 bottom)
                    (gl:vertex-2i 0 top)
                    (gl:vertex-2i (width obj) top)
                    (gl:vertex-2i (width obj) bottom)
                  (gl:end)))
            (if (focussed obj)
                (gl:color-4f 0.5f0 0.5f0 0.5f0 0.5f0)
              (gl:color-4f 0f0 0f0 0f0 0f0))
            (gl:begin gl:+quads+)
              (gl:vertex-2i 0 bottom)
              (gl:vertex-2i 0 top)
              (gl:vertex-2i (width obj) top)
              (gl:vertex-2i (width obj) bottom)
            (gl:end)
            (leave-2d-mode))

              (sgum:free-foreign-object char-advance)))

      (progn
        (sgum:free-foreign-object viewport)
        (sgum:free-foreign-object winx)
        (sgum:free-foreign-object winy)
        (sgum:free-foreign-object winz))))))

(defmethod obj-source ((tx gltext))
  (vmsource tx))

(defmethod obj-sink ((tx gltext))
  (vmsink tx))

; TODO: make this blink
(defun cursor-visible () t)

(defun init-scene (font)
  (let ((red (make-color))
        (blue (make-color))
        (green (make-color)))
    (setf (color-r red) 1f0)
    (setf (color-b blue) 1f0)
    (setf (color-g green) 1f0)
    (let (#|(triangle (make-instance 'glshape
                                   :primitive gl:+triangles+
                                   :vertices '((0f0 1f0 0f0) (-1f0 -1f0 0f0) (1f0 -1f0 0f0))
                                   :color red
                                   :offset (make-pos :x -1.5f0 :z 0f0)))
          (square (make-instance 'glshape 
                                 :primitive gl:+quads+
                                 :vertices '((-1f0 -1f0 0f0) (-1f0 1f0 0f0) (1f0 1f0 0f0) (1f0 -1f0 0f0))
                                 :color (copy-color blue)
                                 :offset (make-pos :x 1.5f0 :z 0f0))) |#
          (sentence (make-gltext "Hello, CLOS." 0 0 0))
          (sentence2 (make-gltext "The quick brown fox jumped over the lazy dog." 0 3 0))
          #|(catobj (make-instance 'glcat
                                 :offset (make-pos :x 10f0 :y 0f0 :z 0f0)
                                 :color (copy-color green))) |#
          )

      ;return the scene object
      (make-instance 'scene :shapes (list #| triangle square catobj |#)
                     :texts (list sentence2 sentence)))))


(defparameter *scene* nil)

(defparameter *mvm* (sgum:allocate-foreign-object gl:double 16))

; given x,y screen coordinates, return x,y,z world coordinates of where that point
; intersects the xy-plane at z=0...this is the plane where all Sync objects live for now.
; returns multiple values (x y)
(defun position-object (x y)
  (let ((winx (coerce x 'double-float))
        (winy (coerce y 'double-float))
        winz
        (modelmatrix *mvm*)
        (projmatrix (sgum:allocate-foreign-object gl:double 16))
        (viewport (sgum:allocate-foreign-object gl:int 4))
        (x1p (sgum:allocate-foreign-object gl:double 1))
        (y1p (sgum:allocate-foreign-object gl:double 1))
        (z1p (sgum:allocate-foreign-object gl:double 1))
        (x2p (sgum:allocate-foreign-object gl:double 1))
        (y2p (sgum:allocate-foreign-object gl:double 1))
        (z2p (sgum:allocate-foreign-object gl:double 1))
        objx objy xslope yslope x1 y1 x2 y2 z1 z2
        realy zdist)

    (unwind-protect
        (progn

          (gl:get-doublev gl:+projection-matrix+ projmatrix)
          (gl:get-integerv gl:+viewport+ viewport)
          (setf realy (- (sgum:deref-array viewport gl:int 3) winy 1))
          (setf winz 0.0d0)
          (glu:unproject winx realy winz modelmatrix projmatrix viewport x1p y1p z1p)
          (setf winz 1.0d0)
          (glu:unproject winx realy winz modelmatrix projmatrix viewport x2p y2p z2p)
          (setf x1 (sgum:deref-pointer x1p gl:double))
          (setf y1 (sgum:deref-pointer y1p gl:double))
          (setf z1 (sgum:deref-pointer z1p gl:double))
          (setf x2 (sgum:deref-pointer x2p gl:double))
          (setf y2 (sgum:deref-pointer y2p gl:double))
          (setf z2 (sgum:deref-pointer z2p gl:double))
          (setf xslope (/ (- x2 x1) (- z2 z1)))
          (setf yslope (/ (- y2 y1) (- z2 z1)))
          (setf objz 0.0d0)
          (setf zdist (- objz z1))
          (setf objx (+ x1 (* xslope zdist)))
          (setf objy (+ y1 (* yslope zdist)))

          (values (coerce objx 'single-float) (coerce objy 'single-float) 0.0f0))

          (progn
            (sgum:free-foreign-object projmatrix)
            (sgum:free-foreign-object viewport)
            (sgum:free-foreign-object x1p)
            (sgum:free-foreign-object y1p)
            (sgum:free-foreign-object z1p)
            (sgum:free-foreign-object x2p)
            (sgum:free-foreign-object y2p)
            (sgum:free-foreign-object z2p)))))
#|

(defun make-triangle-brush (font)
  #'(lambda (x y)
      (multiple-value-bind (objx objy objz)
          (position-object x y)

        (format t "text brush: obj=<~A,~A,~A>~%"
                objx objy objz)

        (make-instance 'glshape
                       :primitive gl:+triangles+
                       :vertices '((0f0 1f0 0f0) (-1f0 -1f0 0f0) (1f0 -1f0 0f0))
                       :color (make-color :r 1f0 :g 0f0 :b 0f0 :a 0f0)
                       :offset (make-pos :x objx :y objy :z objz)))))

; rewrite to use position-object:
(defun make-square-brush (font)
  #'(lambda (x y)
      (let ((winx (coerce x 'double-float))
            (winy (coerce y 'double-float))
            ;(winz 0.06006006006d0)
            ;(winz 0.0d0)
            winz
            (modelmatrix *mvm*)
            (projmatrix (sgum:allocate-foreign-object gl:double 16))
            (viewport (sgum:allocate-foreign-object gl:int 4))
            (x1p (sgum:allocate-foreign-object gl:double 1))
            (y1p (sgum:allocate-foreign-object gl:double 1))
            (z1p (sgum:allocate-foreign-object gl:double 1))
            (x2p (sgum:allocate-foreign-object gl:double 1))
            (y2p (sgum:allocate-foreign-object gl:double 1))
            (z2p (sgum:allocate-foreign-object gl:double 1))
            objx objy xslope yslope x1 y1 x2 y2 z1 z2
            realy zdist)

        (unwind-protect
            (progn

        ;(gl:get-doublev gl:+modelview-matrix+ modelmatrix)
        (gl:get-doublev gl:+projection-matrix+ projmatrix)
        (gl:get-integerv gl:+viewport+ viewport)
        (setf realy (- (sgum:deref-array viewport gl:int 3) winy 1))
        ;(setf realy winy)
        (setf winz 0.0d0)
        (glu:unproject winx realy winz modelmatrix projmatrix viewport x1p y1p z1p)
        (setf winz 1.0d0)
        (glu:unproject winx realy winz modelmatrix projmatrix viewport x2p y2p z2p)
        (setf x1 (sgum:deref-pointer x1p gl:double))
        (setf y1 (sgum:deref-pointer y1p gl:double))
        (setf z1 (sgum:deref-pointer z1p gl:double))
        (setf x2 (sgum:deref-pointer x2p gl:double))
        (setf y2 (sgum:deref-pointer y2p gl:double))
        (setf z2 (sgum:deref-pointer z2p gl:double))
        (format t "near clip point: <~A,~A,~A>~%" x1 y1 z1)
        (format t "far clip point: <~A,~A,~A>~%" x2 y2 z2)
        ;(scene-add-shape *scene* (make-instance 'globj
        ;               :primitive gl:+lines+
        ;               :vertices `((,(coerce x1 'single-float)
        ;                            ,(coerce y1 'single-float)
        ;                            ,(coerce z1 'single-float)) 
        ;                           (,(coerce x2 'single-float)
        ;                            ,(coerce y2 'single-float) 
        ;                            ,(coerce z2 'single-float)))
        ;               :color (make-color :b 1f0)
        ;               :offset (make-pos)
        ;               ))
        
        (setf xslope (/ (- x2 x1) (- z2 z1)))
        (setf yslope (/ (- y2 y1) (- z2 z1)))
        (setf objz 0.0d0)
        (setf zdist (- objz z1))
        (setf objx (+ x1 (* xslope zdist)))
        (setf objy (+ y1 (* yslope zdist)))
        (format t "positioning...obj=<~A,~A,~A>~%"
                objx objy objz)
        ;(make-instance 'globj
        ;               :primitive gl:+triangles+
        ;               :vertices '((0f0 1f0 0f0) (-1f0 -1f0 0f0) (1f0 -1f0 0f0))
        ;               :color (make-color :r 1f0)
        (make-instance 'glshape 
                       :primitive gl:+quads+
                       :vertices '((-1f0 -1f0 0f0) (-1f0 1f0 0f0) (1f0 1f0 0f0) (1f0 -1f0 0f0))
                       :color (make-color :b 1f0)
                       :offset (make-pos :x (coerce objx 'single-float)
                                         :y (coerce objy 'single-float)
                                         :z (coerce objz 'single-float))))
          (progn
            (sgum:free-foreign-object projmatrix)
            (sgum:free-foreign-object viewport)
            (sgum:free-foreign-object x1p)
            (sgum:free-foreign-object y1p)
            (sgum:free-foreign-object z1p)
            (sgum:free-foreign-object x2p)
            (sgum:free-foreign-object y2p)
            (sgum:free-foreign-object z2p))))))
|#

(defun make-text-brush (font)
  #'(lambda (x y)
      (multiple-value-bind (objx objy objz)
          (position-object x y)

        (format t "text brush: obj=<~A,~A,~A>~%"
                objx objy objz)

        (make-gltext "" objx objy objz))))

(defun gltext-callback (sink value)
  (set-text (sink-parent sink) value)
  (scene-changed))

(defmethod initialize-instance :after ((tobj gltext) &rest initargs)
  (setf (sink-parent (vmsink tobj)) tobj))

(defun make-gltext (s x y z)
  (let* ((id (+ 1 (maxdb:lastid *db* 'gltext)))
         (idsink (+ 1 (maxdb:lastid *db* 'sink)))
         (idsource (+ 1 (maxdb:lastid *db* 'source)))
         (mysource (make-instance 'glsource :id idsource :type 'string))
         (mysink (make-instance 'glsink :id idsink :type 'string :callback #'gltext-callback))
         (instance (make-instance 'gltext
                                  :id id
                                  :sink mysink
                                  :source mysource
                                  :color (make-color :r 1f0 :g 1f0 :b 1f0 :a 1f0))))
    (maxdb:type-add-value *gltexttype* :id id)
    (maxdb:function-update *gltext-string* id (max:make-string-address s))
    ; XXX - correct API needs scene to be its own max datatype w/CLOS object class.
    (maxdb:function-update *gltext-scene* id (max:make-number (scene-id *scene*)))
    (maxdb:function-update *gltext-x* id (max:make-number x))
    (maxdb:function-update *gltext-y* id (max:make-number y))
    (maxdb:function-update *gltext-z* id (max:make-number z))
    (maxdb:type-add-value *sink-type* :id idsink)
    (maxdb:function-update *gltext-sink* id (max:make-number idsink))
    (maxdb:type-add-value *source-type* :id idsource)
    (maxdb:function-update *gltext-source* id (max:make-number idsource))

    instance))

(defparameter *brush* nil)

(defmethod start-keyfocus ((obj globj))
  (setf *keyfocus* obj))

(defmethod start-keyfocus ((tobj gltext))
  (if (not (focussed tobj))
      (progn
        (maxdb:rollback *db*)
        (maxdb:begin *db*)))
  (call-next-method))

(defun end-keyfocus ()
  (setf *keyfocus* nil))

(defmethod add-object ((s scene) x y)
  (maxdb:rollback *db*)
  (let ((obj (funcall *brush* x y)))
    (if (eq 'gltext (type-of obj))
        (progn
          (start-keyfocus obj)
          (scene-add-text s obj)
          (setf (selected obj) t)
          ; GL selection ID
          (setf *select* (list (length (shapes s)))))
      (progn
        (scene-add-shape s obj)
        (setf *select* '(0))))))

(defun raster-x ()
  (sgum:deref-array *rp* gl:float 0))
(defun raster-y ()
  (sgum:deref-array *rp* gl:float 1))

(defun draw-scene (scene surface width height font)
  (gl:clear-color 0f0 0f0 0f0 0f0)
  (gl:clear (logior gl:+color-buffer-bit+
                    gl:+depth-buffer-bit+))
  (gl:matrix-mode gl:+modelview+)
  (gl:load-identity)
  (gl:translate-f *x* *y* *z*)
  (gl:rotate-f *xr* 1f0 0f0 0f0)
  (gl:rotate-f *yr* 0f0 1f0 0f0)

  (gl:get-doublev gl:+modelview-matrix+ *mvm*)

  ;(unless (= 0f0 *z*)
  ;  (gl:raster-pos-3f 0f0 0f0 0f0))

  ;(gl:get-floatv gl:+current-raster-position+ *rp*)
  ;(format t "Raster position: ~A~%" (loop for n from 0 to 3 collect (sgum:deref-array *rp* gl:float n)))

  (gl:disable gl:+texture-2d+)

  ; draw shapes first
  (loop for i from 0 to (- (length (drawlist scene)) 1) do
    (progn
      (gl:load-name i)
      (gl:push-matrix)
      (draw (nth i (drawlist scene)) width height)
      (gl:pop-matrix)
      ))
  
  (sdl:gl-swap-buffers)

  ;disable WASD movement since they check "key held" status separately from edit mode
  ;(move)

  t)


(defun print-4x4-matrix (matrix)
  (format t "~%[")
  (loop for i from 0 to 3
    unless (= i 0) do (format t " ")
    do 
    (loop for j from 0 to 3
      do (format t "~A" (sgum:deref-array matrix gl:float (+ i (* j 4))))
      unless (= j 3) do (format t " ")
    )
    when (= i 3) do (format t "]")
    do (format t "~%")))

(defun shift-p ()
  (or (aref *keys* 304)
      (aref *keys* 303)))
(defun printable-p (key)
  (and (>= key sdl:+k-space+) (<= key sdl:+k-z+)))

(defparameter +keymap-shift+
  '((sdl:+k-1+ . #\!)
    (sdl:+k-2+ . #\@)
    (sdl:+k-3+ . #\#)
    (sdl:+k-4+ . #\$)
    (sdl:+k-5+ . #\%)
    (sdl:+k-6+ . #\^)
    (sdl:+k-7+ . #\&)
    (sdl:+k-8+ . #\*)
    (sdl:+k-8+ . #\*)
    (sdl:+k-9+ . #\()
    (sdl:+k-0+ . #\))
    (sdl:+k-minus+ . #\_)
    (sdl:+k-equals+ . #\+)
    (sdl:+k-backslash+ . #\|)
    (sdl:+k-backquote+ . #\~)
    (sdl:+k-comma+ . #\<)
    (sdl:+k-period+ . #\>)
    (sdl:+k-slash+ . #\?)
    (sdl:+k-semicolon+ . #\:)
    (sdl:+k-quote+ . #\")
    (sdl:+k-leftbracket+ . #\{)
    (sdl:+k-rightbracket+ . #\})))

(defun shifted-char (key)
  (cond
   ((and (>= key sdl:+k-a+) (<= key sdl:+k-z+))
    (code-char (- key 32)))
   (t
    (let ((newkey (find key +keymap-shift+ :test #'(lambda (o zi) (eq o (eval (car zi)))))))
      (if newkey
          (cdr newkey)
        (code-char key))))))

(defun modded (mod-state modifier)
    (not (= 0 (logand mod-state modifier))))

(defun handle-edit (textobj key)
  (let* ((ch (code-char key))
         (text (text textobj))
         (len (length text))
         (point (point textobj))
         (left (subseq text 0 point))
         (right (subseq text point len))
         (mod (sdl:get-mod-state))
         (unprintable-modifiers (logand (lognot sdl:+kmod-shift+) mod)))
    (format t "Mod state: ~A~%" mod)
    (cond
     ; don't add character to string when any modifiers except shift are held down
     ((and (printable-p key) (= unprintable-modifiers 0))
      (setf text (concatenate 'string 
                              left 
                              (string (if (shift-p) (shifted-char key) ch))
                              right))
      (incf (point textobj)))
     ((and (eq ch #\Backspace)
           (> point 0))
      (setf text (concatenate 'string (subseq left 0 (- point 1)) right))
      (decf (point textobj)))

     ((eq key sdl:+k-return+)
      (maxdb:commit *db*)
      (end-keyfocus)
      ; delete this text object from scene if text just became empty
      (if (= 0 len) 
          (progn
            (unselect) ; unselect this object, fine to clear selection until multi-select
            (scene-delete-gltext *scene* textobj)
            (return-from handle-edit nil))))

     ((and (eq key sdl:+k-delete+) (< point len))
      (setf text (concatenate 'string left (subseq text (+ point 1) len))))
      
     ((and (> point 0)
           (or (eq key sdl:+k-left+)
               (and (eq key sdl:+k-b+) (modded mod sdl:+kmod-ctrl+))))
      (decf (point textobj)))
     ((and (< point len)
           (or (eq key sdl:+k-right+)
               (and (eq key sdl:+k-f+) (modded mod sdl:+kmod-ctrl+))))
      (incf (point textobj)))
     ((or (eq key sdl:+k-home+)
          (and (eq key sdl:+k-a+) (modded mod sdl:+kmod-ctrl+)))
      (setf (point textobj) 0))
     ((or (eq key sdl:+k-end+)
          (and (eq key sdl:+k-e+) (modded mod sdl:+kmod-ctrl+)))
      (setf (point textobj) len))
     ((and (eq key sdl:+k-u+) (modded mod sdl:+kmod-ctrl+))
      (setf text right)
      (setf (point textobj) 0)))

    (set-text textobj text)))

(defun unselect ()
  (if *select* (setf (selected (selected-object)) nil))
  (setf *select* nil))

(defun selected-object ()
  (if *select* 
      (nth (car *select*) (drawlist *scene*))
    nil))

(defmethod scene-delete-selected ((s scene))
  ;(setf (shapes s) (remove-if #'selected (shapes s)))
  (let ((newshapes nil))
    (dolist (sobj (shapes s))
      (if (selected sobj)
          (obj-delete sobj)
        (push sobj newshapes)))
    (setf (shapes s) newshapes))
  (let ((newtexts nil))
    (dolist (tobj (texts s))
      (if (selected tobj)
          (obj-delete tobj)
        (push tobj newtexts)))
    (setf (texts s) newtexts))
  (let ((newarrows nil))
    (dolist (a (scene-arrows s))
      (if (selected a)
          (obj-delete a)
        (push a newarrows)))
    (setf (scene-arrows s) newarrows))
  (setf *select* nil))

(defmethod scene-delete-gltext ((s scene) (target gltext))
  (let ((newtexts nil))
    (dolist (tobj (texts s))
      (if (eq tobj target)
          (obj-delete tobj)
        (push tobj newtexts)))
    (setf (texts s) newtexts)))

(defun scene-delete-arrows (s targets)
  (let ((newarrows nil))
    (dolist (aobj (scene-arrows s))
      (if (member aobj targets)
          (obj-delete aobj)
        (push aobj newarrows)))
    (setf (scene-arrows s) newarrows)))

(defmethod obj-delete ((tobj gltext))
  (let* ((id (gltext-id tobj))
         (sinkid (maxdb:db-apply *db* *gltext-sink* id))
         (sourceid (maxdb:db-apply *db* *gltext-source* id)))

    (maxdb:function-remove-mapping *gltext-scene* id)
    (maxdb:function-remove-mapping *gltext-x* id)
    (maxdb:function-remove-mapping *gltext-y* id)
    (maxdb:function-remove-mapping *gltext-z* id)
    (maxdb:function-remove-mapping *gltext-string* id)
    (maxdb:function-remove-mapping *gltext-sink* id)
    (maxdb:function-remove-mapping *gltext-source* id)

    (maxdb:type-remove-value *sink-type* :id sinkid)
    (maxdb:type-remove-value *source-type* :id sourceid)

    (maxdb:type-remove-value *gltexttype* :id id)))

(defmethod obj-delete ((a glarrow))
  (maxvmm:delete-pipe (tail-sink a) (head-source a))
  (maxdb:relation-remove-tuple 
   *pipe* (max:make-number (sink-id (tail-sink a))) (max:make-number (source-id (head-source a))))
  (let ((headobj-arrows (obj-arrows (headobj a)))
        (tailobj-arrows (obj-arrows (tailobj a))))
    (obj-delete-arrow (headobj a) a)
    (obj-delete-arrow (tailobj a) a)))

;keyboard focus
(defparameter *keyfocus* nil)
; flag for drag transactions.
(defparameter *drag* nil)

(defun cancel-edit ()
  (end-keyfocus)
  (maxdb:rollback *db*)
  (scene-changed))

(defun event-loop (surface w h bpp video-flags fps font)
  (declare (ignorable surface))
  (declare (optimize (debug 3)))
  (sdl:event-loop
   (:expose ()
            (format t "Expose event~%")
            (scene-changed))
   (:active (gain) ; cl-sdl bug, documented to have 2 args but only 1 works
            ; redraw scene if window becomes active
            (if gain (scene-changed)))
   (:key-down (key)
              (setf (aref *keys* key) t)
              (format t "Key down: ~A~%" key)

              (cond
               ; Esc always clears selection
               ;((= key (char-code #\Esc))
               ((= key sdl:+k-escape+)
                (cancel-edit)
                (unselect))

               ((and (not (null *keyfocus*))
                     (eq 'gltext (type-of *keyfocus*)))
                (handle-edit *keyfocus* key)
                (scene-changed))

               ;press b: change brush to a bool (actually a square)
               #| ((= key sdl:+k-b+)
                (setf *brush* (make-square-brush *font*))) |#
               ((= key sdl:+k-c+)
                (setf *brush* (make-cat-brush *font*)))
               ((= key sdl:+k-t+)
                (setf *brush* (make-text-brush *font*)))
               ((= key sdl:+k-m+)
                  (let ((matrix (sgum:allocate-foreign-object gl:float 16)))
                    (unwind-protect
                        (progn
                          (gl:get-floatv gl:+modelview-matrix+ matrix)
                          (format t "Model view matrix: ")
                          (print-4x4-matrix matrix)
                          (format t "Stored *mvm*: ")
                          (print-4x4-matrix matrix))
                      (sgum:free-foreign-object matrix))))
              ((= key sdl:+k-p+)
                  (let ((matrix (sgum:allocate-foreign-object gl:float 16)))
                    (unwind-protect
                        (progn
                          (gl:get-floatv gl:+projection-matrix+ matrix)
                          (format t "Projection matrix (before 2D): ")
                          (print-4x4-matrix matrix))
                      (sgum:free-foreign-object matrix))))
              ((= key sdl:+k-q+)
                  (return))
              ((= key sdl:+k-r+)
                  (format t "Raster position: ~A~%" (loop for n from 0 to 3 collect (sgum:deref-array *rp* gl:float n))) 
                  )
              ((= key sdl:+k-delete+)
               (scene-delete-selected *scene*)
               (scene-changed))))
   (:key-up (key)
            (setf (aref *keys* key) nil))
            ;(format t "Key up: ~A~%" key))
   (:mouse-button-up (button x y)
                     ;(format t "Mouse button ~A up at (~A, ~A)~%" button x y)
                     (setf (aref *mb* button) nil)
                     (if *drag*
                         (progn
                           (maxdb:commit *db*)
                           (setf *drag* nil))))
   (:mouse-button-down (button x y)
                       (format t "Mouse button ~A down at (~A, ~A)~%" button x y)
                       (setf (aref *mb* button) t)
                       (if (= button 5)
                           (progn
                             (setf *z* (- *z* *dz*))
                             (scene-changed)))
                       (if (= button 4)
                           (progn
                             (setf *z* (+ *z* *dz*))
                             (scene-changed)))
                       (if (= button 1)
                           (let ((old-selected-object (selected-object))
                                 (old-selection *select*)
                                 (mod (sdl:get-mod-state)))
                           ; need to re-render scene now that it's not drawn continually to prevent flicker during select? why?
                             (draw-scene *scene* surface w h font) 
                             (unselect)
                             (select surface w h font)
                             (if (modded mod sdl:+kmod-alt+)
                                 (progn
                                   (make-arrow old-selected-object (selected-object))
                                   (setf *select* old-selection)
                                   (setf (selected old-selected-object) t))
                               (progn
                                 (if (null *select*)
                                     (add-object *scene* x y))
                                 (handle-select (selected-object) (cdr *select*) x y)))
                             (scene-changed))))
   (:mouse-motion (x y xrel yrel state)
                  ;(format t "Mouse motion: x=~A y=~A xrel=~A yrel=~A state=~A~%" x y xrel yrel state)
                  (setf *mx* x)
                  (setf *my* y)
                  (if (aref *mb* 3)
                      (progn
                        (incf *x* (* *dxm* xrel))
                        (incf *y* (* *dym* yrel))
                        (scene-changed)))
                  (if (aref *mb* 2)
                      (progn
                        ; Moving mouse left<->right rotates around Y axis and vice versa
                        (incf *xr* (* *dxr* yrel))
                        (incf *yr* (* *dyr* xrel))
                        (scene-changed)))
                  (if (aref *mb* 1)
                      ; if there is a selected object
                      (if (and *select* (not (modded (sdl:get-mod-state) sdl:+kmod-alt+)))
                          (multiple-value-bind (pointerx pointery)
                              (position-object x y)
                            (let ((obj (selected-object)))
                              (if (draggable obj)
                                  (let* ((distancex (abs (- x *clickx*)))
                                         (distancey (abs (- y *clicky*))))
                                    (format t "distancex=~A, distancey=~A~%" distancex distancey)
                                    (if (and (not *drag*) (or (> distancex *drag-threshold*)
                                                              (> distancey *drag-threshold*)))
                                        (progn ; begin drag
                                          (cancel-edit)
                                          (setf *drag* t)
                                          (maxdb:begin *db*)))
                                    (if *drag* (progn ; drag the selected object
                                                 (setf (obj-x obj) (- pointerx (dragx obj)))
                                                 (setf (obj-y obj) (- pointery (dragy obj)))
                                                 (scene-changed))))))))))
   (:quit ()
          (return))
   (:resize (width height)
            (setf w width)
            (setf h height)
            (format t "Resized width = ~A height = ~A~%" width height)
            (sdl:free-surface surface)
            (setf surface (sdl:set-video-mode width height bpp video-flags))
            (resize surface width height)
            (gl:matrix-mode gl:+projection+)
            (gl:get-doublev gl:+projection-matrix+ *proj-presel*) 
            (scene-changed))
   (:idle ()
          (if *scene-changed* 
              (progn 
                (format t "idle, width=~A height=~A~%" w h)
                (draw-scene *scene* surface w h font)
                (incf-fps fps)
                (setf *scene-changed* nil)))
          (loop while (maxvmm:queue-run-next))
          ; whether or not scene changed, sleep 10ms
          (sdl:delay 10))))

(defparameter *scene-changed* t)
(defun scene-changed ()
  (setf *scene-changed* t))

(defparameter *font* nil)

(defun scroll (&optional datastore)
  (let ((width 1024)
        (height 768)
        (video-flags (logior sdl:+opengl+ sdl:+resizable+ sdl:+gl-doublebuffer+ sdl:+hwpalette+))
        (bpp 16)
        fps
        font)

  (sdl:init sdl:+init-video+) 

  (multiple-value-bind (hw-available-p blit-hw-p)
      (sdl:query-video-info :hw-available :blit-hw)
    (setf video-flags (logior video-flags
                              (if hw-available-p
                                  sdl:+hwsurface+
                                  sdl:+swsurface+)
                              (if blit-hw-p
                                  sdl:+hwaccel+
                                 0))))

  (sdl:gl-set-attribute sdl:+gl-doublebuffer+ 1)

  #| to create a maximized (but not fullscreen) window,
     trick SDL by requesting fullscreen modes, then using
     that resolution.  If requesting mode list without
     specifying sdl:+fullscreen+ flag, -1 is returned,
     so this is the only way to detect screen size. |#
  (let* ((modes (sdl:list-modes nil sdl:+fullscreen+))
         (firstmode (sgum:deref-array modes (* sdl:rect) 0))
         (firstrect (sgum:deref-pointer firstmode sdl:rect)))
    (setf width (sdl:rect-w firstrect))
    (setf height (sdl:rect-h firstrect)))
  (let ((surface (sdl:set-video-mode width height bpp video-flags)))
    (when (sgum:null-pointer-p surface)
      (error "Failed to obtain surface"))

    (gl:enable gl:+texture-2d+)
    (gl:shade-model gl:+smooth+)
    (gl:clear-color 0.0f0 0.0f0 0.0f0 0.0f0)
    (gl:clear-depth 1.0d0)
    (gl:enable gl:+depth-test+)
    (gl:depth-func gl:+lequal+)
    (gl:hint gl:+perspective-correction-hint+ gl:+nicest+)

    (resize surface width height)

    (setf *font* (init-font))

    (sdl:wm-set-caption "max" "max")
    (let ((scenelist (maxdb:quotient *db* *scene-name* datastore)))
      (setf *scene*
            (if (null scenelist)
                (new-scene datastore)
              (load-scene (car scenelist)))))
          
    (scene-changed); draw scene- if window starts in foreground, no expose event.

    (setf *brush* (make-text-brush *font*))

    (gl:matrix-mode gl:+projection+)
    (gl:get-doublev gl:+projection-matrix+ *proj-presel*)

    (sdl:enable-key-repeat sdl:+default-repeat-delay+ sdl:+default-repeat-interval+)

    ; start FPS counter
    (setf fps (make-fps))

    (unwind-protect (event-loop surface width height bpp video-flags fps *font*)
      (sdl:quit))
    (values))))

;(scroll)

(defun load-scene (n)
  (let ((s (make-instance 'scene :id n))
        (textids (maxdb:quotient *db* *gltext-scene* n))
        (catids (maxdb:quotient *db* *concatenator-scene* n))
        (sink-id-to-instance (make-hash-table)); there's now a method on class glsink named sink-parent. it's different.
        (source-id-to-instance (make-hash-table))
        (source-parent (make-hash-table))
        (blue (make-color :b 1f0)))
    (dolist (textid textids)
      (let* ((sinkid (maxdb:db-apply *db* *gltext-sink* textid))
             (sourceid (maxdb:db-apply *db* *gltext-source* textid))
             (sink (make-instance 'glsink :id sinkid :type 'string :callback #'gltext-callback))
             (source (make-instance 'glsource :id sourceid :type 'string))
             (tobj (make-instance 'gltext :id textid :sink sink :source source)))

; to load pipes, first index all text objs and catobjs by sink ID and source ID (sink-parent, source-parent)
; then go thru index, for all source IDs, query the pipe relation to find the relevant pipes.
; (skip sinks since they will all be accounted for, iterating over sinks would make each pipe twice)
; for each pipe, create glarrow object with head/tail linked to appropriate glcat/gltext.
        (setf (gethash sinkid sink-id-to-instance) sink)
        (setf (gethash sourceid source-id-to-instance) source)
        (setf (gethash sourceid source-parent) tobj)
        (push tobj (texts s))))
    (dolist (catid catids)
      (let* ((idsinkl (maxdb:db-apply *db* *concatenator-inputl* catid))
             (idsinkr (maxdb:db-apply *db* *concatenator-inputr* catid))
             (idsource (maxdb:db-apply *db* *concatenator-output* catid))
             (mysinkl (make-instance 'funcom-sink :id idsinkl :type 'string :callback #'glcat-callback-left))
             (mysinkr (make-instance 'funcom-sink :id idsinkr :type 'string :callback #'glcat-callback-right))
             (mysource (make-instance 'glsource :id idsource :type 'string))
             (cobj (make-instance 'glcat :glcat-id catid 
                                  :inputl mysinkl :inputr mysinkr
                                  :output mysource
                                  :color (make-color :r 0f0 :g 1f0 :b 0f0 :a 1f0))))
           
        (setf (gethash idsinkl sink-id-to-instance) mysinkl)
        (setf (gethash idsinkr sink-id-to-instance) mysinkr)
        (setf (gethash idsource source-id-to-instance) mysource)
        (setf (gethash idsource source-parent) cobj)
        (push cobj (shapes s))))
    ; load arrows
    (maphash #'(lambda (sourceid sourceobj)
                 (let ((sinks (maxdb:relation-queryr *pipe* (max:make-number sourceid))))
                   (dolist (sinkid sinks)
                     (let* ((tailobj (sink-parent (gethash sinkid sink-id-to-instance)))
                            (sink (gethash sinkid sink-id-to-instance))
                            (source (gethash sourceid source-id-to-instance))
                            (a (make-instance 'glarrow :color blue
                                              :tailobj tailobj
                                              :headobj sourceobj
                                              :tail-sink sink
                                              :head-source source)))
                       (scene-add-arrow s a)
                       (push a (obj-arrows tailobj))
                       (push a (obj-arrows sourceobj))
                       (maxvmm:make-pipe sink source)))))
             source-parent)

    ; send initial updates from all text sources to prep component local value caches.
    (mapcar #'(lambda (tobj)
                (maxvmm:update (vmsource tobj) (text tobj)))
            (texts s))

    s))

(defun new-scene (name)
  (let* ((id (+ 1 (maxdb:lastid *db* 'scene)))
        (s (make-instance 'scene :id id)))
    (maxdb:type-add-value *scene-type* :id id)
    (maxdb:function-update *scene-name* id (max:make-string-address name))
    s))
