#| Max Virtual Machine Monitor: A data flow engine |#

(in-package :maxvmm)

#| class source: represents the producer of a stream of discrete
values.  Each source has a type, and every value originated by the
source is a member of that type. |#

(defclass source ()
  ((ty :initform nil :initarg :type :accessor source-type)
   (value :initform nil :initarg :value :accessor source-value :documentation "most recent value")
   (sinks :initform '() :initarg :sinks :accessor source-sinks :documentation "list of sinks depending on this source")))

#| class sink: represents a consumer of a stream of discrete values.
Each sink has 0 or 1 sources of the same type (pipe: connection from a
sink to a source, a data conduit).  Each sink has a callback LISP
function for the VMM to notify the sink's owner of new data
arriving. |#

(defclass sink ()
  ((ty :initform nil :initarg :type :accessor sink-type)
   (source :initform nil :initarg :source :accessor sink-source)
   (callback :initform nil :initarg :callback :accessor sink-callback)))

#| *RUNQ*: a FIFO queue implemented as a regular list.  
   (car head) indicates the next pipe to run |#

(defclass runqueue ()
  ((head :initform nil :initarg :head :accessor q-head)))

(defparameter *runq* (make-instance 'runqueue))

#| Method UPDATE: The source's owner calls UPDATE to provide new data
from this source.  Adds this job to the queue immediately.  When
processed, expand the job to multiple jobs for each of this source's
dependent sinks. |#

(defmethod update ((src source) value)
  (setf (source-value src) value)
  (queue-add *runq* src value))

(defmethod run ((src source) value)
  (dolist (sink (source-sinks src))
    (queue-add *runq* sink value))
  src)

(defmethod run ((snk sink) value)
  (funcall (sink-callback snk) snk value)
  snk)

(defclass job ()
  ((object :initarg :object :accessor job-object)
   (value :initarg :value :accessor job-value)))

(defmethod queue-add ((r runqueue) item value)
  (setf (q-head r) (nconc (q-head r) (cons (make-instance 'job 
                                                          :object item 
                                                          :value value) 
                                           '())))
  item)

#| RUN-NEXT-PIPE: takes a single job from the queue and runs it.
Running a job will cause callbacks to its sink, which may result in
additional work being added to the queue. |#

(defun queue-run-next ()
  (unless (null (q-head *runq*))
    (let* ((job (pop (q-head *runq*)))
           (object (job-object job))
           (value (job-value job)))
      (run object value))))

#| MAKE-PIPE: register a dependency - pipe - from a source to a sink.
Also runs the sink immediately to notify it of the source's current
value. |#

(defun make-pipe (sink source)
  (assert (eq (source-type source) (sink-type sink)))
  (setf (sink-source sink) source)
  (push sink (source-sinks source))
  (run sink (source-value source)))

(defun delete-pipe (sink source)
  (setf (sink-source sink) nil)
  (setf (source-sinks source) 
        (delete sink (source-sinks source))))

; TODO: pause
; TODO: stop
